# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING, Literal, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.scipy as jsp

from ott import utils
from ott.geometry import geometry, low_rank, pointcloud
from ott.problems.linear import linear_problem
from ott.problems.quadratic import quadratic_costs
from ott.types import Transport

if TYPE_CHECKING:
  from ott.solvers.linear import sinkhorn_lr

__all__ = ["QuadraticProblem"]


@jax.tree_util.register_pytree_node_class
class QuadraticProblem:
  r"""Quadratic OT problem.

  The quadratic loss of a single OT matrix is assumed to
  have the form given in :cite:`peyre:16`, eq. 4.

  The two geometries below parameterize matrices :math:`C` and :math:`\bar{C}`
  in that equation. The function :math:`L` (of two real values) in that equation
  is assumed to match the form given in eq. 5., with our notations:

  .. math::
    L(x, y) = f_1(x) + f_2(y) - h_1(x) h_2(y)

  Args:
    geom_xx: Ground geometry of the first space.
    geom_yy: Ground geometry of the second space.
    geom_xy: Geometry defining the linear penalty term for
      fused Gromov-Wasserstein :cite:`vayer:19`. If :obj:`None`, the problem
      reduces to a plain Gromov-Wasserstein problem :cite:`peyre:16`.
    fused_penalty: Multiplier of the linear term in fused Gromov-Wasserstein,
      i.e. ``problem = purely quadratic + fused_penalty * linear problem``.
    scale_cost: How to rescale the cost matrices. If a :class:`str`,
      use specific options available in :class:`~ott.geometry.geometry.Geometry`
      or :class:`~ott.geometry.pointcloud.PointCloud`. If :obj:`None`, keep
      the original scaling.
    a: The first marginal. If :obj:`None`, it will be uniform.
    b: The second marginal. If :obj:`None`, it will be uniform.
    loss: Gromov-Wasserstein loss function, see
      :class:`~ott.problems.quadratic.quadratic_costs.GWLoss` for more
      information.
    tau_a: If :math:`< 1.0`, defines how much unbalanced the problem is on
      the first marginal.
    tau_b: If :math:`< 1.0`, defines how much unbalanced the problem is on
      the second marginal.
    gw_unbalanced_correction: Whether the unbalanced version of
      :cite:`sejourne:21` is used. Otherwise, ``tau_a`` and ``tau_b``
      only affect the inner Sinkhorn loop.
    ranks: Ranks of the cost matrices, see
      :meth:`~ott.geometry.geometry.Geometry.to_LRCGeometry`. Used when
      geometries are *not* :class:`~ott.geometry.pointcloud.PointCloud` with
      `'sqeucl'` cost function. If `-1`, the geometries will not be converted
      to low-rank. If :class:`tuple`, it specifies the ranks of ``geom_xx``,
      ``geom_yy`` and ``geom_xy``, respectively. If :class:`int`, rank is shared
      across all geometries.
    tolerances: Tolerances used when converting geometries to low-rank. Used
      when geometries are not :class:`~ott.geometry.pointcloud.PointCloud` with
      `'sqeucl'` cost. If :class:`float`, it is shared across all geometries.
  """

  def __init__(
      self,
      geom_xx: geometry.Geometry,
      geom_yy: geometry.Geometry,
      geom_xy: Optional[geometry.Geometry] = None,
      fused_penalty: float = 1.0,
      scale_cost: Optional[Union[float, str]] = None,
      a: Optional[jnp.ndarray] = None,
      b: Optional[jnp.ndarray] = None,
      loss: Union[Literal["sqeucl", "kl"], quadratic_costs.GWLoss] = "sqeucl",
      tau_a: float = 1.0,
      tau_b: float = 1.0,
      gw_unbalanced_correction: bool = True,
      ranks: Union[int, Tuple[int, ...]] = -1,
      tolerances: Union[float, Tuple[float, ...]] = 1e-2,
  ):
    if scale_cost is not None:
      geom_xx = geom_xx.set_scale_cost(scale_cost)
      geom_yy = geom_yy.set_scale_cost(scale_cost)
      if geom_xy is not None:
        geom_xy = geom_xy.set_scale_cost(scale_cost)

    self._geom_xx = geom_xx
    self._geom_yy = geom_yy
    self._geom_xy = geom_xy
    self.fused_penalty = fused_penalty
    self.scale_cost = scale_cost
    self._a = a
    self._b = b
    self.tau_a = tau_a
    self.tau_b = tau_b
    self.gw_unbalanced_correction = gw_unbalanced_correction
    self.ranks = ranks
    self.tolerances = tolerances

    self._loss_name = loss
    if self._loss_name == "sqeucl":
      self.loss = quadratic_costs.make_square_loss()
    elif loss == "kl":
      self.loss = quadratic_costs.make_kl_loss()
    else:
      self.loss = loss

  def marginal_dependent_cost(
      self,
      marginal_1: jnp.ndarray,
      marginal_2: jnp.ndarray,
  ) -> low_rank.LRCGeometry:
    r"""Initialize cost term that depends on the marginals of the transport.

    Uses the first term in eq. 6, p. 1 of :cite:`peyre:16`.

    Let :math:`p` be the `[n,]` marginal of the transport matrix for samples
    from :attr:`geom_xx` and :math:`q` the `[m,]` marginal of the
    transport matrix for samples from :attr:`geom_yy`.

    When ``cost_xx`` (resp. ``cost_yy``) is the cost matrix of :attr:`geom_xx`
    (resp. :attr:`geom_yy`), the cost term that depends on these marginals can
    be written as:

    .. math::

      \text{marginal_dep_term} = \text{lin1}(\text{cost_xx}) p \mathbb{1}_{m}^T
                      +  \mathbb{1}_{n}(\text{lin2}(\text{cost_yy}) q)^T

    This helper function instantiates these two low-rank matrices and groups
    them into a single low-rank cost geometry object.

    Args:
      marginal_1: [n,], first marginal of transport matrix.
      marginal_2: [m,], second marginal of transport matrix.

    Returns:
      Low-rank geometry of rank 2, storing normalization constants.
    """
    geom_xx, geom_yy = self.geom_xx, self.geom_yy
    if self._loss_name == "sqeucl":  # quadratic apply, efficient for LR
      tmp1 = geom_xx.apply_square_cost(marginal_1, axis=1)
      tmp2 = geom_yy.apply_square_cost(marginal_2, axis=1)
    else:
      f1, f2 = self.linear_loss
      tmp1 = apply_cost(geom_xx, marginal_1, axis=1, fn=f1)
      tmp2 = apply_cost(geom_yy, marginal_2, axis=1, fn=f2)
    x_term = jnp.stack([tmp1, jnp.ones_like(tmp1)], axis=-1)
    y_term = jnp.stack([jnp.ones_like(tmp2), tmp2], axis=-1)
    return low_rank.LRCGeometry(cost_1=x_term, cost_2=y_term)

  def cost_unbalanced_correction(
      self,
      transport_matrix: jnp.ndarray,
      marginal_1: jnp.ndarray,
      marginal_2: jnp.ndarray,
      epsilon: float,
  ) -> float:
    r"""Calculate cost term from the quadratic divergence when unbalanced.

    In the unbalanced setting (``tau_a < 1.0 or tau_b < 1.0``), the
    introduction of a quadratic divergence :cite:`sejourne:21` adds a term
    to the GW local cost.

    Let :math:`a` [num_a,] be the target weights for samples
    from geom_xx and :math:`b` [num_b,] be the target weights
    for samples from `geom_yy`. Let :math:`P` [num_a, num_b] be the transport
    matrix, :math:`P1` the first marginal and :math:`P^T1` the second marginal.
    The term of the cost matrix coming from the quadratic KL in the
    unbalanced case can be written as:

    `unbalanced_correction_term` =
        :math:`tau_a / (1 - tau_a) * \sum(KL(P1|a))`
        :math:`+ tau_b / (1 - tau_b) * \sum(KL(P^T1|b))`
        :math:`+ epsilon * \sum(KL(P|ab'))`

    Args:
      transport_matrix: jnp.ndarray<float>[num_a, num_b], transport matrix.
      marginal_1: jnp.ndarray<float>[num_a,], marginal of the transport matrix
        for samples from :attr:`geom_xx`.
      marginal_2: jnp.ndarray<float>[num_b,], marginal of the transport matrix
        for samples from :attr:`geom_yy`.
      epsilon: entropy regularizer.

    Returns:
      The cost term.
    """

    def regularizer(tau: float) -> float:
      return epsilon * tau / (1.0 - tau)

    marginal_1loga = jsp.special.xlogy(marginal_1, self.a).sum()
    marginal_2logb = jsp.special.xlogy(marginal_2, self.b).sum()

    cost = epsilon * jsp.special.xlogy(transport_matrix, transport_matrix).sum()
    if self.tau_a != 1.0:
      cost += regularizer(
          self.tau_a
      ) * (-jsp.special.entr(marginal_1).sum() - marginal_1loga)
    if self.tau_b != 1.0:
      cost += regularizer(
          self.tau_b
      ) * (-jsp.special.entr(marginal_2).sum() - marginal_2logb)
    return cost

  # TODO(michalk8): highly coupled to the pre-defined initializer, refactor
  def init_transport_mass(self) -> float:
    """Initialize the transport mass.

    Returns:
      The sum of the elements of the normalized transport matrix.
    """
    a = jax.lax.stop_gradient(self.a)
    b = jax.lax.stop_gradient(self.b)
    return a.sum() * b.sum()

  def update_lr_geom(
      self,
      lr_sink: "sinkhorn_lr.LRSinkhornOutput",
      relative_epsilon: Optional[bool] = None,
  ) -> geometry.Geometry:
    """Recompute (possibly LRC) linearization using LR Sinkhorn output."""
    marginal_1 = lr_sink.marginal(1)
    marginal_2 = lr_sink.marginal(0)
    marginal_cost = self.marginal_dependent_cost(marginal_1, marginal_2)

    # Extract factors from LR Sinkhorn output
    q, r, inv_sqg = lr_sink.q, lr_sink.r, 1.0 / jnp.sqrt(lr_sink.g)
    # Distribute middle marginal evenly across both factors.
    q, r = q * inv_sqg[None, :], r * inv_sqg[None, :]

    # Handle LRC Geometry case.
    h1, h2 = self.quad_loss
    geom_xx, geom_yy, geom_xy = self.geom_xx, self.geom_yy, self.geom_xy
    tmp1 = apply_cost(geom_xx, q, axis=1, fn=h1)
    tmp2 = apply_cost(geom_yy, r, axis=1, fn=h2)
    if self.is_low_rank:
      geom = low_rank.LRCGeometry(
          cost_1=tmp1, cost_2=-tmp2, relative_epsilon=relative_epsilon
      ) + marginal_cost
      if self.is_fused:
        geom = geom + geom_xy
    else:
      cost_matrix = marginal_cost.cost_matrix - jnp.dot(tmp1, tmp2.T)
      cost_matrix += self.fused_penalty * self._fused_cost_matrix
      geom = geometry.Geometry(
          cost_matrix=cost_matrix, relative_epsilon=relative_epsilon
      )
    return geom  # noqa: RET504

  def update_linearization(
      self,
      transport: Transport,
      epsilon: Optional[float] = None,
      old_transport_mass: float = 1.0,
      relative_epsilon: Optional[bool] = None,
  ) -> linear_problem.LinearProblem:
    """Update linearization of GW problem by updating cost matrix.

    If the problem is balanced (``tau_a = 1.0 and tau_b = 1.0``), the equation
    follows eq. 6, p. 1 of :cite:`peyre:16`.

    If the problem is unbalanced (``tau_a < 1.0 or tau_b < 1.0``), two cases are
    possible, as explained in :meth:`init_linearization` above.
    Finally, it is also possible to consider a Fused Gromov-Wasserstein problem.
    Details about the resulting cost matrix are also given in
    :meth:`init_linearization`.

    Args:
      transport: Solution of the linearization of the quadratic problem.
      epsilon: An epsilon scheduler or a float passed on to the linearization.
      old_transport_mass: Sum of the elements of the transport matrix at the
        previous iteration.
      relative_epsilon: Whether to use relative epsilon in the linearized
        geometry.

    Returns:
      Updated linear OT problem, a new local linearization of GW problem.
    """
    rescale_factor = 1.0
    unbalanced_correction = 0.0

    if not self.is_balanced:
      marginal_1 = transport.marginal(axis=1)
      transport_mass = jax.lax.stop_gradient(marginal_1.sum())
      rescale_factor = jnp.sqrt(old_transport_mass / transport_mass)

    marginal_1 = transport.marginal(axis=1) * rescale_factor
    marginal_2 = transport.marginal(axis=0) * rescale_factor
    marginal_cost = self.marginal_dependent_cost(marginal_1, marginal_2)

    transport_matrix = transport.matrix * rescale_factor

    if not self.is_balanced:
      epsilon *= jax.lax.stop_gradient(marginal_1.sum())
      unbalanced_correction = self.cost_unbalanced_correction(
          transport_matrix, marginal_1, marginal_2, epsilon=epsilon
      )

    h1, h2 = self.quad_loss
    geom_xx, geom_yy = self.geom_xx, self.geom_yy

    tmp = apply_cost(geom_xx, transport_matrix, axis=1, fn=h1)
    tmp = apply_cost(geom_yy, tmp.T, axis=1, fn=h2).T

    cost_matrix = marginal_cost.cost_matrix - tmp + unbalanced_correction
    cost_matrix += self.fused_penalty * rescale_factor * self._fused_cost_matrix

    geom = geometry.Geometry(
        cost_matrix=cost_matrix,
        epsilon=epsilon,
        relative_epsilon=relative_epsilon,
    )

    return linear_problem.LinearProblem(
        geom, self.a, self.b, tau_a=self.tau_a, tau_b=self.tau_b
    )

  def update_lr_linearization(
      self,
      lr_sink: "sinkhorn_lr.LRSinkhornOutput",
      *,
      relative_epsilon: Optional[bool] = None,
  ) -> linear_problem.LinearProblem:
    """Update a Quad problem linearization using a LR Sinkhorn."""
    return linear_problem.LinearProblem(
        self.update_lr_geom(lr_sink, relative_epsilon=relative_epsilon),
        self.a,
        self.b,
        tau_a=self.tau_a,
        tau_b=self.tau_b
    )

  @property
  def _fused_cost_matrix(self) -> Union[float, jnp.ndarray]:
    return self.geom_xy.cost_matrix if self.is_fused else 0.0

  @property
  def _is_low_rank_convertible(self) -> bool:

    def convertible(geom: geometry.Geometry) -> bool:
      return isinstance(geom, low_rank.LRCGeometry) or (
          isinstance(geom, pointcloud.PointCloud) and geom.is_squared_euclidean
      )

    if self.is_low_rank:
      return True

    geom_xx, geom_yy, geom_xy = self.geom_xx, self.geom_yy, self.geom_xy
    # either explicitly via cost factorization or implicitly (e.g., a PC)
    return self.ranks != -1 or (
        convertible(geom_xx) and convertible(geom_yy) and
        (geom_xy is None or convertible(geom_xy))
    )

  def to_low_rank(
      self,
      rng: Optional[jax.Array] = None,
  ) -> "QuadraticProblem":
    """Convert geometries to low-rank.

    Args:
      rng: Random key for seeding.

    Returns:
      Quadratic problem with low-rank geometries.
    """

    def convert(
        vals: Union[int, float, Tuple[Union[int, float], ...]]
    ) -> Tuple[Union[int, float], ...]:
      size = 2 + self.is_fused
      if isinstance(vals, (int, float)):
        return (vals,) * 3
      assert len(vals) == size, vals
      return vals + (None,) * (3 - size)

    if self.is_low_rank:
      return self

    rng = utils.default_prng_key(rng)
    rng1, rng2, rng3 = jax.random.split(rng, 3)
    (geom_xx, geom_yy, geom_xy, *children), aux_data = self.tree_flatten()
    (r1, r2, r3), (t1, t2, t3) = convert(self.ranks), convert(self.tolerances)

    geom_xx = geom_xx.to_LRCGeometry(rank=r1, tol=t1, rng=rng1)
    geom_yy = geom_yy.to_LRCGeometry(rank=r2, tol=t2, rng=rng2)
    if self.is_fused:
      if isinstance(
          geom_xy, pointcloud.PointCloud
      ) and geom_xy.is_squared_euclidean:
        geom_xy = geom_xy.to_LRCGeometry()
      else:
        geom_xy = geom_xy.to_LRCGeometry(rank=r3, tol=t3, rng=rng3)

    return type(self).tree_unflatten(
        aux_data, [geom_xx, geom_yy, geom_xy] + children
    )

  @property
  def geom_xx(self) -> geometry.Geometry:
    """Geometry of the first space."""
    return self._geom_xx

  @property
  def geom_yy(self) -> geometry.Geometry:
    """Geometry of the second space."""
    return self._geom_yy

  @property
  def geom_xy(self) -> Optional[geometry.Geometry]:
    """Geometry of the joint space."""
    return self._geom_xy

  @property
  def a(self) -> jnp.ndarray:
    """First marginal."""
    num_a = self.geom_xx.shape[0]
    return jnp.ones((num_a,)) / num_a if self._a is None else self._a

  @property
  def b(self) -> jnp.ndarray:
    """Second marginal."""
    num_b = self.geom_yy.shape[0]
    return jnp.ones((num_b,)) / num_b if self._b is None else self._b

  @property
  def is_fused(self) -> bool:
    """Whether the problem is fused."""
    return self.geom_xy is not None

  @property
  def is_low_rank(self) -> bool:
    """Whether all geometries are low-rank."""
    return (
        isinstance(self.geom_xx, low_rank.LRCGeometry) and
        isinstance(self.geom_yy, low_rank.LRCGeometry) and
        (not self.is_fused or isinstance(self.geom_xy, low_rank.LRCGeometry))
    )

  @property
  def linear_loss(self) -> Tuple[quadratic_costs.Loss, quadratic_costs.Loss]:
    """Linear part of the Gromov-Wasserstein loss."""
    return self.loss.f1, self.loss.f2

  @property
  def quad_loss(self) -> Tuple[quadratic_costs.Loss, quadratic_costs.Loss]:
    """Quadratic part of the Gromov-Wasserstein loss."""
    return self.loss.h1, self.loss.h2

  @property
  def is_balanced(self) -> bool:
    """Whether the problem is balanced."""
    return ((not self.gw_unbalanced_correction) or
            (self.tau_a == 1.0 and self.tau_b == 1.0))

  def tree_flatten(self):  # noqa: D102
    return ([self.geom_xx, self.geom_yy, self.geom_xy, self._a, self._b], {
        "tau_a": self.tau_a,
        "tau_b": self.tau_b,
        "loss": self._loss_name,
        "fused_penalty": self.fused_penalty,
        "scale_cost": self.scale_cost,
        "gw_unbalanced_correction": self.gw_unbalanced_correction,
        "ranks": self.ranks,
        "tolerances": self.tolerances
    })

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    geoms, (a, b) = children[:3], children[3:]
    return cls(*geoms, a=a, b=b, **aux_data)


def apply_cost(  # noqa: D103
    geom: geometry.Geometry, arr: jnp.ndarray, *, axis: int,
    fn: quadratic_costs.Loss
) -> jnp.ndarray:
  return geom.apply_cost(arr, axis=axis, fn=fn.func, is_linear=fn.is_linear)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Optional, Tuple

import jax
import jax.numpy as jnp

from flax import linen as nn

__all__ = ["PositiveDense", "PosDefPotentials"]

PRNGKey = jax.Array
Shape = Tuple[int, ...]
Dtype = Any
Array = jnp.ndarray

DEFAULT_KERNEL_INIT = lambda *a, **k: nn.initializers.lecun_normal()(*a, **k)
DEFAULT_BIAS_INIT = nn.initializers.zeros
DEFAULT_RECTIFIER = nn.activation.relu


class PositiveDense(nn.Module):
  """A linear transformation using a matrix with all entries non-negative.

  Args:
    dim_hidden: Number of output dimensions.
    rectifier_fn: Rectifier function. The default is
      :func:`~flax.linen.activation.relu`.
    use_bias: Whether to add bias to the output.
    kernel_init: Initializer for the matrix. The default is
      :func:`~flax.linen.initializers.lecun_normal`.
    bias_init: Initializer for the bias. The default is
      :func:`~flax.linen.initializers.zeros`.
    precision: Numerical precision of the computation.
  """

  dim_hidden: int
  rectifier_fn: Callable[[Array], Array] = DEFAULT_RECTIFIER
  use_bias: bool = True
  kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = DEFAULT_KERNEL_INIT
  bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = DEFAULT_BIAS_INIT
  precision: Optional[jax.lax.Precision] = None

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
    """Applies a linear transformation to x along the last dimension.

    Args:
      x: Array of shape ``[batch, ..., features]``.

    Returns:
      Array of shape ``[batch, ..., dim_hidden]``.
    """
    # TODO(michalk8): update when refactoring neuraldual
    # assert x.ndim > 1, x.ndim

    kernel = self.param(
        "kernel", self.kernel_init, (x.shape[-1], self.dim_hidden)
    )
    kernel = self.rectifier_fn(kernel)

    x = jnp.tensordot(x, kernel, axes=(-1, 0), precision=self.precision)
    if self.use_bias:
      x = x + self.param("bias", self.bias_init, (self.dim_hidden,))

    return x


class PosDefPotentials(nn.Module):
  r""":math:`\frac{1}{2} x^T (A_i A_i^T + \text{Diag}(d_i)) x + b_i^T x^2 + c_i`
    potentials.

  This class implements a layer that takes (batched) ``d``-dimensional vectors
  ``x`` in, to output a ``num_potentials``-dimensional vector. Each of the
  entries in that output is a positive definite quadratic form evaluated at
  ``x``; each of these quadratic terms is parameterized as a low-rank plus
  diagonal matrix. The low-rank term is parameterized as :math:`A_i A_i^T`,
  where each of these matrices is of size ``(rank, d)``. Taken together,
  these matrices form a tensor ``(num_potentials, rank, d)``.
  The diagonal terms :math:`d_i` form a ``(num_potentials, d)`` matrix of
  positive values; the linear terms :math:`b_i` form a ``(num_potentials, d)``
  matrix. Finally, the :math:`c_i` are contained in a vector of size
  ``(num_potentials,)``.

  Args:
    num_potentials: Dimension of the output.
    rank: Rank of the matrices :math:`A_i` used as low-rank factors
      for the quadratic potentials.
    rectifier_fn: Rectifier function to ensure non-negativity of the diagonals
      :math:`d_i`. The default is :func:`~flax.linen.activation.relu`.
    use_linear: Whether to add a linear layers :math:`b_i` to the outputs.
    use_bias: Whether to add biases :math:`c_i` to the outputs.
    kernel_lr_init: Initializer for the matrices :math:`A_i`
      of the quadratic potentials when ``rank > 0``.
      The default is :func:`~flax.linen.initializers.lecun_normal`.
    kernel_diag_init: Initializer for the diagonals :math:`d_i`.
      The default is :func:`~flax.linen.initializers.ones`.
    kernel_linear_init: Initializer for the linear layers :math:`b_i`.
      The default is :func:`~flax.linen.initializers.lecun_normal`.
    bias_init: Initializer for the bias. The default is
      :func:`~flax.linen.initializers.zeros`.
    precision: Numerical precision of the computation.
  """  # noqa: D205,E501

  num_potentials: int
  rank: int = 0
  rectifier_fn: Callable[[Array], Array] = DEFAULT_RECTIFIER
  use_linear: bool = True
  use_bias: bool = True
  kernel_lr_init: Callable[[PRNGKey, Shape, Dtype], Array] = DEFAULT_KERNEL_INIT
  kernel_diag_init: Callable[[PRNGKey, Shape, Dtype],
                             Array] = nn.initializers.ones
  kernel_linear_init: Callable[[PRNGKey, Shape, Dtype],
                               Array] = DEFAULT_KERNEL_INIT
  bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = DEFAULT_BIAS_INIT
  precision: Optional[jax.lax.Precision] = None

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
    """Compute quadratic forms of the input.

    Args:
      x: Array of shape ``[batch, ..., features]``.

    Returns:
      Array of shape ``[batch, ..., num_potentials]``.
    """
    # TODO(michalk8): update when refactoring neuraldual
    # assert x.ndim > 1, x.ndim

    dim_data = x.shape[-1]
    x = x.reshape((-1, dim_data))

    diag_kernel = self.param(
        "diag_kernel", self.kernel_diag_init, (dim_data, self.num_potentials)
    )
    # ensures the diag_kernel parameter stays non negative
    diag_kernel = self.rectifier_fn(diag_kernel)

    # (batch, dim_data, 1), (1, dim_data, num_potentials)
    y = 0.5 * jnp.sum(((x ** 2)[..., None] * diag_kernel[None]), axis=1)

    if self.rank > 0:
      quad_kernel = self.param(
          "quad_kernel", self.kernel_lr_init,
          (self.num_potentials, dim_data, self.rank)
      )
      # (batch, num_potentials, rank)
      quad = 0.5 * jnp.tensordot(
          x, quad_kernel, axes=(-1, 1), precision=self.precision
      ) ** 2
      y = y + jnp.sum(quad, axis=-1)

    if self.use_linear:
      linear_kernel = self.param(
          "lin_kernel", self.kernel_linear_init,
          (dim_data, self.num_potentials)
      )
      y = y + jnp.dot(x, linear_kernel, precision=self.precision)

    if self.use_bias:
      y = y + self.param("bias", self.bias_init, (self.num_potentials,))

    return y

  @classmethod
  def init_from_samples(
      cls, source: jnp.ndarray, target: jnp.ndarray, **kwargs: Any
  ) -> "PosDefPotentials":
    """Initialize the layer using Gaussian approximation :cite:`bunne:22`.

    Args:
      source: Samples from the source distribution, array of shape ``[n, d]``.
      target: Samples from the target distribution, array of shape ``[m, d]``.
      kwargs: Keyword arguments for initialization. Note that ``use_linear``
        will be always set to :obj:`True`.

    Returns:
      The layer with fixed linear and quadratic initialization.
    """
    factor, mean = _compute_gaussian_map_params(source, target)

    kwargs["use_linear"] = True
    return cls(
        kernel_lr_init=lambda *_, **__: factor,
        kernel_linear_init=lambda *_, **__: mean.T,
        **kwargs,
    )


def _compute_gaussian_map_params(
    source: jnp.ndarray, target: jnp.ndarray
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  from ott.math import matrix_square_root
  from ott.tools.gaussian_mixture import gaussian

  g_s = gaussian.Gaussian.from_samples(source)
  g_t = gaussian.Gaussian.from_samples(target)
  lin_op = g_s.scale.gaussian_map(g_t.scale)
  b = jnp.squeeze(g_t.loc) - lin_op @ jnp.squeeze(g_s.loc)
  lin_op = matrix_square_root.sqrtm_only(lin_op)

  return jnp.expand_dims(lin_op, 0), jnp.expand_dims(b, 0)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from types import MappingProxyType
from typing import Any, Mapping, Optional, Tuple

import jax.numpy as jnp

from ott.geometry import costs, pointcloud, segment
from ott.problems.linear import linear_problem
from ott.solvers.linear import sinkhorn


def segment_sinkhorn(
    x: jnp.ndarray,
    y: jnp.ndarray,
    num_segments: Optional[int] = None,
    max_measure_size: Optional[int] = None,
    cost_fn: Optional[costs.CostFn] = None,
    segment_ids_x: Optional[jnp.ndarray] = None,
    segment_ids_y: Optional[jnp.ndarray] = None,
    indices_are_sorted: bool = False,
    num_per_segment_x: Optional[Tuple[int, ...]] = None,
    num_per_segment_y: Optional[Tuple[int, ...]] = None,
    weights_x: Optional[jnp.ndarray] = None,
    weights_y: Optional[jnp.ndarray] = None,
    sinkhorn_kwargs: Mapping[str, Any] = MappingProxyType({}),
    **kwargs: Any
) -> jnp.ndarray:
  """Compute regularized OT cost between subsets of vectors in ``x`` and ``y``.

  Helper function designed to compute Sinkhorn regularized OT cost between
  several point clouds of varying size, in parallel, using padding.
  In practice, The inputs ``x`` and ``y`` (and their weight vectors `weights_x`
  and ``weights_y``) are assumed to be large weighted point clouds, that
  describe points taken from multiple measures. To extract several subsets of
  points, we provide two interfaces. The first interface assumes that a vector
  of id's is passed, describing for each point of ``x`` (resp. ``y``) to which
  measure the point belongs to. The second interface assumes that ``x`` and
  ``y`` were simply formed by concatenating several measures contiguously, and
  that only indices that segment these groups are needed to recover them.

  For both interfaces, both ``x`` and ``y`` should contain the same total number
  of segments. Each segment will be padded as necessary, all segments rearranged
  as a tensor, and :func:`jax.vmap` used to evaluate Sinkhorn divergences in
  parallel.

  Args:
    x: Array of input points, of shape ``[num_x, dim]``.
      Multiple segments are held in this single array.
    y: Array of target points, of shape ``[num_y, dim]``.
    num_segments: Number of segments contained in ``x``and ``y``.
      Providing this is required for JIT compilation to work,
      see also :func:`~ott.geometry.segment.segment_point_cloud`.
    max_measure_size: Total size of measures after padding. Should ideally be
      set to an upper bound on points clouds processed with the segment
      interface. Providing this is required for JIT compilation to work.
    cost_fn: Cost function, defaults to
      :class:`~ott.geometry.costs.SqEuclidean`.
    segment_ids_x: **1st interface** The segment ID for which each row of `x`
      belongs. This is a similar interface to `jax.ops.segment_sum`.
    segment_ids_y: **1st interface** The segment ID for which each row of `y`
      belongs.
    indices_are_sorted: **1st interface** Whether `segment_ids_x` and
      `segment_ids_y` are sorted.
    num_per_segment_x: **2nd interface** Number of points in each segment in
      `x`. For example, ``[100, 20, 30]`` would imply that ``x``is segmented
      into 3 arrays of length ``[100]``, ``[20]``, and ``[30]`` respectively.
    num_per_segment_y: **2nd interface** Number of points in each segment in
      ``y``.
    weights_x: Weights of each input points, arranged in the same segmented
      order as ``x``.
    weights_y: Weights of each input points, arranged in the same segmented
      order as ``y``.
    sinkhorn_kwargs: Optionally a dict containing the keywords arguments for
      calls for the :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver,
      called three times to evaluate for each segment the Sinkhorn regularized
      OT cost between ``x``/``y``, ``x``/`x`, and `y`/``y`` (except when
      ``static_b`` is ``True``, in which case ``y``/``y`` is not evaluated).
    kwargs: keywords arguments passed to form
      :class:`~ott.geometry.pointcloud.PointCloud` geometry objects from the
      subsets of points and masses selected in ``x``and ``y``, possibly a
      :class:`~ott.geometry.costs.CostFn` or an entropy regularizer.

  Returns:
    An array of Sinkhorn regularized OT costs for each segment.
  """
  # instantiate padding vector
  dim = x.shape[1]
  if cost_fn is None:
    # default padder
    padding_vector = costs.CostFn._padder(dim=dim)
  else:
    padding_vector = cost_fn._padder(dim=dim)

  def eval_fn(
      padded_x: jnp.ndarray,
      padded_y: jnp.ndarray,
      padded_weight_x: jnp.ndarray,
      padded_weight_y: jnp.ndarray,
  ) -> float:
    mask_x = padded_weight_x > 0.0
    mask_y = padded_weight_y > 0.0

    geom = pointcloud.PointCloud(
        padded_x,
        padded_y,
        cost_fn=cost_fn,
        src_mask=mask_x,
        tgt_mask=mask_y,
        **kwargs,
    )
    prob = linear_problem.LinearProblem(
        geom, a=padded_weight_x, b=padded_weight_y
    )
    solver = sinkhorn.Sinkhorn(**sinkhorn_kwargs)
    return solver(prob).reg_ot_cost

  return segment._segment_interface(
      x,
      y,
      eval_fn,
      num_segments=num_segments,
      max_measure_size=max_measure_size,
      segment_ids_x=segment_ids_x,
      segment_ids_y=segment_ids_y,
      indices_are_sorted=indices_are_sorted,
      num_per_segment_x=num_per_segment_x,
      num_per_segment_y=num_per_segment_y,
      weights_x=weights_x,
      weights_y=weights_y,
      padding_vector=padding_vector
  )



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Any,
    Callable,
    Dict,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import jax
import jax.numpy as jnp
import numpy as np

from ott.geometry import geometry
from ott.initializers.quadratic import initializers as quad_initializers
from ott.math import fixed_point_loop
from ott.problems.linear import linear_problem
from ott.problems.quadratic import quadratic_problem
from ott.solvers import was_solver
from ott.solvers.linear import sinkhorn, sinkhorn_lr

__all__ = ["GromovWasserstein", "GWOutput"]

LinearOutput = Union[sinkhorn.SinkhornOutput, sinkhorn_lr.LRSinkhornOutput]

ProgressCallbackFn = Callable[
    [Tuple[np.ndarray, np.ndarray, np.ndarray, "GWState"]], None]


class GWOutput(NamedTuple):
  """Holds the output of the Gromov-Wasserstein solver.

  Args:
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    converged: Convergence flag for the outer GW iterations.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
    linear_state: State used to solve and store solutions to the local
      linearization of GW.
    geom: The geometry underlying the local linearization.
    old_transport_mass: Holds total mass of transport at previous iteration.
  """

  costs: Optional[jnp.ndarray] = None
  linear_convergence: Optional[jnp.ndarray] = None
  converged: bool = False
  errors: Optional[jnp.ndarray] = None
  linear_state: Optional[LinearOutput] = None
  geom: Optional[geometry.Geometry] = None
  # Intermediate values.
  old_transport_mass: float = 1.0

  def set(self, **kwargs: Any) -> "GWOutput":
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  @property
  def matrix(self) -> jnp.ndarray:
    """Transport matrix."""
    return self._rescale_factor * self.linear_state.matrix

  def apply(self, inputs: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Apply the transport to an array; axis=1 for its transpose."""
    return self._rescale_factor * self.linear_state.apply(inputs, axis=axis)

  @property
  def reg_gw_cost(self) -> float:
    """Regularized optimal transport cost of the linearization."""
    return self.linear_state.reg_ot_cost

  @property
  def _rescale_factor(self) -> float:
    return jnp.sqrt(self.old_transport_mass / self.linear_state.transport_mass)

  @property
  def primal_cost(self) -> float:
    """Return transport cost of current linear OT solution at geometry."""
    return self.linear_state.transport_cost_at_geom(other_geom=self.geom)

  @property
  def n_iters(self) -> int:  # noqa: D102
    if self.errors is None:
      return -1
    return jnp.sum(self.errors[:, 0] != -1)


class GWState(NamedTuple):
  """State of the Gromov-Wasserstein solver.

  Attributes:
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    linear_state: State used to solve and store solutions to the local
      linearization of GW.
    linear_pb: Local linearization of the quadratic GW problem.
    old_transport_mass: Intermediary value of the mass of the transport matrix.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
  """

  costs: jnp.ndarray
  linear_convergence: jnp.ndarray
  linear_state: LinearOutput
  linear_pb: linear_problem.LinearProblem
  old_transport_mass: float
  errors: Optional[jnp.ndarray] = None

  def set(self, **kwargs: Any) -> "GWState":
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  def update(  # noqa: D102
      self, iteration: int, linear_sol: LinearOutput,
      linear_pb: linear_problem.LinearProblem, store_errors: bool,
      old_transport_mass: float
  ) -> "GWState":
    costs = self.costs.at[iteration].set(linear_sol.reg_ot_cost)
    errors = None
    if store_errors and self.errors is not None:
      errors = self.errors.at[iteration, :].set(linear_sol.errors)
    linear_convergence = self.linear_convergence.at[iteration].set(
        linear_sol.converged
    )

    return self.set(
        linear_state=linear_sol,
        linear_pb=linear_pb,
        costs=costs,
        linear_convergence=linear_convergence,
        errors=errors,
        old_transport_mass=old_transport_mass
    )


@jax.tree_util.register_pytree_node_class
class GromovWasserstein(was_solver.WassersteinSolver):
  """Gromov-Wasserstein solver :cite:`peyre:16`.

  .. seealso::
    Low-rank Gromov-Wasserstein :cite:`scetbon:23` is implemented in
    :class:`~ott.solvers.quadratic.gromov_wasserstein_lr.LRGromovWasserstein`.

  Args:
    linear_solver: Linear OT solver.
    epsilon: Entropic regularization.
    relative_epsilon: Whether to use relative epsilon in the linearized
      geometry.
    initializer: Quadratic initializer. If :obj:`None`, use
      :class:`~ott.initializers.quadratic.initializers.QuadraticInitializer`
    warm_start: Whether to initialize Sinkhorn calls with the values
      from the previous iteration.
    progress_fn: callback function which gets called during the
      Gromov-Wasserstein iterations, so the user can display the error at each
      iteration, e.g., using a progress bar.
      See :func:`~ott.utils.default_progress_fn` for a basic implementation.
    kwargs: Keyword arguments for
      :class:`~ott.solvers.was_solver.WassersteinSolver`.
  """

  def __init__(
      self,
      linear_solver: sinkhorn.Sinkhorn,
      epsilon: float = 1.0,
      relative_epsilon: Optional[bool] = None,
      initializer: Optional[quad_initializers.BaseQuadraticInitializer] = None,
      warm_start: bool = False,
      progress_fn: Optional[ProgressCallbackFn] = None,
      **kwargs: Any
  ):
    super().__init__(linear_solver, **kwargs)
    self.epsilon = epsilon
    self.relative_epsilon = relative_epsilon
    self.initializer = initializer
    self.warm_start = warm_start
    self.progress_fn = progress_fn

  def __call__(
      self,
      prob: quadratic_problem.QuadraticProblem,
      init: Optional[linear_problem.LinearProblem] = None,
      **kwargs: Any,
  ) -> GWOutput:
    """Run the Gromov-Wasserstein solver.

    Args:
      prob: Quadratic OT problem.
      init: Initial linearization of the quadratic problem. If `None`, it will
        be computed using the initializer.
      kwargs: Keyword arguments used when calling the initializer.

    Returns:
      The Gromov-Wasserstein output.
    """
    if prob._is_low_rank_convertible:
      prob = prob.to_low_rank()

    if init is None:
      initializer = self.initializer
      if initializer is None:
        initializer = quad_initializers.QuadraticInitializer()
      init = initializer(
          prob,
          epsilon=self.epsilon,
          relative_epsilon=self.relative_epsilon,
          **kwargs
      )

    out = iterations(self, prob, init)
    # TODO(lpapaxanthoos): remove stop_gradient when using backprop
    linearization = prob.update_linearization(
        jax.lax.stop_gradient(out.linear_state),
        epsilon=self.epsilon,
        old_transport_mass=jax.lax.stop_gradient(out.old_transport_mass),
        relative_epsilon=self.relative_epsilon,
    )

    linear_state = out.linear_state.set_cost(linearization, True, True)
    iteration = jnp.sum(out.costs != -1)
    converged = jnp.logical_and(
        iteration < self.max_iterations,
        jnp.nanmean(out.linear_convergence) == 1.0
    )
    return out.set(
        linear_state=linear_state, geom=linearization.geom, converged=converged
    )

  def init_state(
      self,
      prob: quadratic_problem.QuadraticProblem,
      init: linear_problem.LinearProblem,
  ) -> GWState:
    """Initialize the state of the Gromov-Wasserstein iterations.

    Args:
      prob: Quadratic OT problem.
      init: Initial linearization of the quadratic problem.

    Returns:
      The initial Gromov-Wasserstein state.
    """
    linear_state = self.linear_solver(init)
    num_iter = self.max_iterations
    transport_mass = prob.init_transport_mass()
    if self.store_inner_errors:
      errors = -jnp.ones((num_iter, self.linear_solver.outer_iterations))
    else:
      errors = None

    return GWState(
        costs=-jnp.ones((num_iter,)),
        linear_convergence=jnp.full((num_iter,), fill_value=jnp.nan),
        linear_state=linear_state,
        linear_pb=init,
        old_transport_mass=transport_mass,
        errors=errors,
    )

  def output_from_state(
      self,
      state: GWState,
  ) -> GWOutput:
    """Create an output from a loop state.

    Arguments:
      state: A GWState.

    Returns:
      A GWOutput.
    """
    return GWOutput(
        costs=state.costs,
        linear_convergence=state.linear_convergence,
        errors=state.errors,
        linear_state=state.linear_state,
        geom=state.linear_pb.geom,
        old_transport_mass=state.old_transport_mass
    )

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    children, aux_data = super().tree_flatten()
    aux_data["epsilon"] = self.epsilon
    aux_data["relative_epsilon"] = self.relative_epsilon
    aux_data["initializer"] = self.initializer
    aux_data["warm_start"] = self.warm_start
    aux_data["progress_fn"] = self.progress_fn
    return children, aux_data

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "GromovWasserstein":
    linear_solver, threshold = children
    return cls(linear_solver, threshold=threshold, **aux_data)


def iterations(
    solver: GromovWasserstein,
    prob: quadratic_problem.QuadraticProblem,
    init: linear_problem.LinearProblem,
) -> GWOutput:
  """Jittable Gromov-Wasserstein outer loop."""

  def cond_fn(
      iteration: int, solver: GromovWasserstein, state: GWState
  ) -> bool:
    return solver._continue(state, iteration)

  def body_fn(
      iteration: int, solver: GromovWasserstein, state: GWState,
      compute_error: bool
  ) -> GWState:
    del compute_error  # always assumed true for the outer loop of GW

    lin_state = state.linear_state
    init = (lin_state.f, lin_state.g) if solver.warm_start else (None, None)
    linear_pb = prob.update_linearization(
        lin_state,
        solver.epsilon,
        state.old_transport_mass,
        relative_epsilon=solver.relative_epsilon,
    )
    out = solver.linear_solver(linear_pb, init=init)

    old_transport_mass = jax.lax.stop_gradient(
        state.linear_state.transport_mass
    )
    new_state = state.update(
        iteration, out, linear_pb, solver.store_inner_errors, old_transport_mass
    )

    # Inner iterations is currently fixed to 1.
    inner_iterations = 1
    if solver.progress_fn is not None:
      jax.debug.callback(
          solver.progress_fn,
          (iteration, inner_iterations, solver.max_iterations, state)
      )

    return new_state

  state = fixed_point_loop.fixpoint_iter(
      cond_fn=cond_fn,
      body_fn=body_fn,
      min_iterations=solver.min_iterations,
      max_iterations=solver.max_iterations,
      inner_iterations=1,
      constants=solver,
      state=solver.init_state(prob, init)
  )

  return solver.output_from_state(state)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, Literal, Optional, Union

import jax.numpy as jnp

from ott.geometry import geometry
from ott.problems.quadratic import quadratic_costs, quadratic_problem
from ott.solvers.linear import sinkhorn
from ott.solvers.quadratic import gromov_wasserstein as gw
from ott.solvers.quadratic import gromov_wasserstein_lr as lrgw

__all__ = ["solve"]


def solve(
    geom_xx: geometry.Geometry,
    geom_yy: geometry.Geometry,
    geom_xy: Optional[geometry.Geometry] = None,
    fused_penalty: float = 1.0,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    tau_a: float = 1.0,
    tau_b: float = 1.0,
    loss: Union[Literal["sqeucl", "kl"], quadratic_costs.GWLoss] = "sqeucl",
    gw_unbalanced_correction: bool = True,
    rank: int = -1,
    linear_solver_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> Union[gw.GWOutput, lrgw.LRGWOutput]:
  """Solve quadratic regularized OT problem using a Gromov-Wasserstein solver.

  Args:
    geom_xx: Ground geometry of the first space.
    geom_yy: Ground geometry of the second space.
    geom_xy: Geometry defining the linear penalty term for
      fused Gromov-Wasserstein :cite:`vayer:19`. If :obj:`None`, the problem
      reduces to a plain Gromov-Wasserstein problem :cite:`peyre:16`.
    fused_penalty: Multiplier of the linear term in fused Gromov-Wasserstein,
      i.e. ``problem = purely quadratic + fused_penalty * linear problem``.
    a: The first marginal. If :obj:`None`, it will be uniform.
    b: The second marginal. If :obj:`None`, it will be uniform.
    tau_a: If :math:`< 1`, defines how much unbalanced the problem is
      on the first marginal.
    tau_b: If :math:`< 1`, defines how much unbalanced the problem is
      on the second marginal.
    loss: Gromov-Wasserstein loss function, see
      :class:`~ott.problems.quadratic.quadratic_costs.GWLoss` for more
      information. If ``rank > 0``, ``'sqeucl'`` is always used.
    gw_unbalanced_correction: Whether the unbalanced version of
      :cite:`sejourne:21` is used. Otherwise, ``tau_a`` and ``tau_b``
      only affect the resolution of the linearization of the GW problem
      in the inner loop. Only used when ``rank = -1``.
    rank: Rank constraint on the coupling to minimize the quadratic OT problem
      :cite:`scetbon:22`. If :math:`-1`, no rank constraint is used.
    linear_solver_kwargs: Keyword arguments for
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn`, if ``rank > 0``.
    kwargs: Keyword arguments for
      :class:`~ott.solvers.quadratic.gromov_wasserstein.GromovWasserstein` or
      :class:`~ott.solvers.quadratic.gromov_wasserstein_lr.LRGromovWasserstein`,
      depending on the ``rank``

  Returns:
    The Gromov-Wasserstein output.
  """
  prob = quadratic_problem.QuadraticProblem(
      geom_xx=geom_xx,
      geom_yy=geom_yy,
      geom_xy=geom_xy,
      fused_penalty=fused_penalty,
      a=a,
      b=b,
      tau_a=tau_a,
      tau_b=tau_b,
      loss=loss,
      gw_unbalanced_correction=gw_unbalanced_correction
  )

  if rank > 0:
    solver = lrgw.LRGromovWasserstein(rank, **kwargs)
  else:
    if linear_solver_kwargs is None:
      linear_solver_kwargs = {}
    linear_solver = sinkhorn.Sinkhorn(**linear_solver_kwargs)
    solver = gw.GromovWasserstein(linear_solver, **kwargs)

  return solver(prob)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import NamedTuple, Optional, Tuple

import jax
import jax.numpy as jnp
import jax.scipy as jsp

from ott.math import fixed_point_loop
from ott.problems.linear import linear_problem

__all__ = ["unbalanced_dykstra_lse", "unbalanced_dykstra_kernel"]


class State(NamedTuple):  # noqa: D101
  v1: jnp.ndarray
  v2: jnp.ndarray
  u1: jnp.ndarray
  u2: jnp.ndarray
  g: jnp.ndarray
  err: float


class Constants(NamedTuple):  # noqa: D101
  a: jnp.ndarray
  b: jnp.ndarray
  rho_a: float
  rho_b: float
  supp_a: Optional[jnp.ndarray] = None
  supp_b: Optional[jnp.ndarray] = None


def unbalanced_dykstra_lse(
    c_q: jnp.ndarray,
    c_r: jnp.ndarray,
    c_g: jnp.ndarray,
    gamma: float,
    ot_prob: linear_problem.LinearProblem,
    translation_invariant: bool = True,
    tolerance: float = 1e-3,
    min_iter: int = 0,
    inner_iter: int = 10,
    max_iter: int = 10000
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
  """Dykstra's algorithm for the unbalanced
  :class:`~ott.solvers.linear.sinkhorn_lr.LRSinkhorn` in LSE mode.

  Args:
    c_q: Cost associated with :math:`Q`.
    c_r: Cost associated with :math:`R`.
    c_g: Cost associated with :math:`g`.
    gamma: The (inverse of) the gradient step.
    ot_prob: Unbalanced OT problem.
    translation_invariant: Whether to use the translation invariant objective,
      see :cite:`scetbon:23`, alg. 3.
    tolerance: Convergence threshold.
    min_iter: Minimum number of iterations.
    inner_iter: Compute error every ``inner_iter``.
    max_iter: Maximum number of iterations.

  Returns:
    The :math:`Q`, :math:`R` and :math:`g` factors.
  """  # noqa: D205

  def _softm(
      v: jnp.ndarray,
      c: jnp.ndarray,
      axis: int,
  ) -> jnp.ndarray:
    v = jnp.expand_dims(v, axis=1 - axis)
    return jsp.special.logsumexp(v + c, axis=axis)

  def _error(
      gamma: float,
      new_state: State,
      old_state: State,
  ) -> float:
    u1_err = jnp.linalg.norm(new_state.u1 - old_state.u1, ord=jnp.inf)
    u2_err = jnp.linalg.norm(new_state.u2 - old_state.u2, ord=jnp.inf)
    v1_err = jnp.linalg.norm(new_state.v1 - old_state.v1, ord=jnp.inf)
    v2_err = jnp.linalg.norm(new_state.v2 - old_state.v2, ord=jnp.inf)
    return (1.0 / gamma) * jnp.max(jnp.array([u1_err, u2_err, v1_err, v2_err]))

  def cond_fn(
      iteration: int,
      const: Constants,
      state: State,
  ) -> bool:
    del iteration, const
    return tolerance < state.err

  def body_fn(
      iteration: int, const: Constants, state: State, compute_error: bool
  ) -> State:
    log_a, log_b = jnp.log(const.a), jnp.log(const.b)
    rho_a, rho_b = const.rho_a, const.rho_b

    c_a = _get_ratio(const.rho_a, gamma)
    c_b = _get_ratio(const.rho_b, gamma)

    if translation_invariant:
      lam_a, lam_b = compute_lambdas(const, state, gamma, g=c_g, lse_mode=True)

      u1 = c_a * (log_a - _softm(state.v1, c_q, axis=1))
      u1 = u1 - lam_a / ((1.0 / gamma) + rho_a)
      u2 = c_b * (log_b - _softm(state.v2, c_r, axis=1))
      u2 = u2 - lam_b / ((1.0 / gamma) + rho_b)

      state_lam = State(
          v1=state.v1, v2=state.v2, u1=u1, u2=u2, g=state.g, err=state.err
      )
      lam_a, lam_b = compute_lambdas(
          const, state_lam, gamma, g=c_g, lse_mode=True
      )

      v1_trans = _softm(u1, c_q, axis=0)
      v2_trans = _softm(u2, c_r, axis=0)

      g_trans = gamma * (lam_a + lam_b) + c_g
    else:
      u1 = c_a * (log_a - _softm(state.v1, c_q, axis=1))
      u2 = c_b * (log_b - _softm(state.v2, c_r, axis=1))

      v1_trans = _softm(u1, c_q, axis=0)
      v2_trans = _softm(u2, c_r, axis=0)
      g_trans = c_g

    g = (1.0 / 3.0) * (g_trans + v1_trans + v2_trans)
    v1 = g - v1_trans
    v2 = g - v2_trans

    new_state = State(v1=v1, v2=v2, u1=u1, u2=u2, g=g, err=jnp.inf)
    err = jax.lax.cond(
        jnp.logical_and(compute_error, iteration >= min_iter),
        _error,
        lambda *_: state.err,
        gamma,
        new_state,
        state,
    )
    return State(v1=v1, v2=v2, u1=u1, u2=u2, g=g, err=err)

  n, m, r = c_q.shape[0], c_r.shape[0], c_g.shape[0]
  constants = Constants(
      a=ot_prob.a,
      b=ot_prob.b,
      rho_a=_rho(ot_prob.tau_a),
      rho_b=_rho(ot_prob.tau_b),
      supp_a=ot_prob.a > 0,
      supp_b=ot_prob.b > 0,
  )
  init_state = State(
      v1=jnp.zeros(r),
      v2=jnp.zeros(r),
      u1=jnp.zeros(n),
      u2=jnp.zeros(m),
      g=c_g,
      err=jnp.inf,
  )

  state: State = fixed_point_loop.fixpoint_iter_backprop(
      cond_fn, body_fn, min_iter, max_iter, inner_iter, constants, init_state
  )

  q = jnp.exp(state.u1[:, None] + c_q + state.v1[None, :])
  r = jnp.exp(state.u2[:, None] + c_r + state.v2[None, :])
  g = jnp.exp(state.g)

  return q, r, g


def unbalanced_dykstra_kernel(
    k_q: jnp.ndarray,
    k_r: jnp.ndarray,
    k_g: jnp.ndarray,
    gamma: float,
    ot_prob: linear_problem.LinearProblem,
    translation_invariant: bool = True,
    tolerance: float = 1e-3,
    min_iter: int = 0,
    inner_iter: int = 10,
    max_iter: int = 10000
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
  """Dykstra's algorithm for the unbalanced
  :class:`~ott.solvers.linear.sinkhorn_lr.LRSinkhorn` in kernel mode.

  Args:
    k_q: Kernel associated with :math:`Q`.
    k_r: Kernel associated with :math:`R`.
    k_g: Kernel associated with :math:`g`.
    gamma: The (inverse of) the gradient step.
    ot_prob: Unbalanced OT problem.
    translation_invariant: Whether to use the translation invariant objective,
      see :cite:`scetbon:23`, alg. 3.
    tolerance: Convergence threshold.
    min_iter: Minimum number of iterations.
    inner_iter: Compute error every ``inner_iter``.
    max_iter: Maximum number of iterations.

  Returns:
    The :math:`Q`, :math:`R` and :math:`g` factors.
  """  # noqa: D205

  def _error(
      gamma: float,
      new_state: State,
      old_state: State,
  ) -> float:
    u1_err = jnp.linalg.norm(
        jnp.log(new_state.u1) - jnp.log(old_state.u1), ord=jnp.inf
    )
    u2_err = jnp.linalg.norm(
        jnp.log(new_state.u2) - jnp.log(old_state.u2), ord=jnp.inf
    )
    v1_err = jnp.linalg.norm(
        jnp.log(new_state.v1) - jnp.log(old_state.v1), ord=jnp.inf
    )
    v2_err = jnp.linalg.norm(
        jnp.log(new_state.v2) - jnp.log(old_state.v2), ord=jnp.inf
    )
    return (1.0 / gamma) * jnp.max(jnp.array([u1_err, u2_err, v1_err, v2_err]))

  def cond_fn(
      iteration: int,
      const: Constants,
      state: State,
  ) -> bool:
    del iteration, const
    return tolerance < state.err

  def body_fn(
      iteration: int, const: Constants, state: State, compute_error: bool
  ) -> State:
    c_a = _get_ratio(const.rho_a, gamma)
    c_b = _get_ratio(const.rho_b, gamma)

    if translation_invariant:
      lam_a, lam_b = compute_lambdas(const, state, gamma, g=k_g, lse_mode=False)

      u1 = jnp.where(const.supp_a, (const.a / (k_q @ state.v1)) ** c_a, 0.0)
      u1 = u1 * jnp.exp(-lam_a / ((1.0 / gamma) + const.rho_a))
      u2 = jnp.where(const.supp_b, (const.b / (k_r @ state.v2)) ** c_b, 0.0)
      u2 = u2 * jnp.exp(-lam_b / ((1.0 / gamma) + const.rho_b))

      state_lam = State(
          v1=state.v1, v2=state.v2, u1=u1, u2=u2, g=state.g, err=state.err
      )
      lam_a, lam_b = compute_lambdas(
          const, state_lam, gamma, g=k_g, lse_mode=False
      )

      v1_trans = k_q.T @ u1
      v2_trans = k_r.T @ u2

      k_trans = jnp.exp(gamma * (lam_a + lam_b)) * k_g
      g = (k_trans * v1_trans * v2_trans) ** (1.0 / 3.0)
    else:
      u1 = jnp.where(const.supp_a, (const.a / (k_q @ state.v1)) ** c_a, 0.0)
      u2 = jnp.where(const.supp_b, (const.b / (k_r @ state.v2)) ** c_b, 0.0)

      v1_trans = k_q.T @ u1
      v2_trans = k_r.T @ u2

      g = (k_g * v1_trans * v2_trans) ** (1.0 / 3.0)

    v1 = g / v1_trans
    v2 = g / v2_trans

    new_state = State(v1=v1, v2=v2, u1=u1, u2=u2, g=g, err=jnp.inf)
    err = jax.lax.cond(
        jnp.logical_and(compute_error, iteration >= min_iter),
        _error,
        lambda *_: state.err,
        gamma,
        new_state,
        state,
    )
    return State(v1=v1, v2=v2, u1=u1, u2=u2, g=g, err=err)

  n, m, r = k_q.shape[0], k_r.shape[0], k_g.shape[0]
  constants = Constants(
      a=ot_prob.a,
      b=ot_prob.b,
      rho_a=_rho(ot_prob.tau_a),
      rho_b=_rho(ot_prob.tau_b),
      supp_a=ot_prob.a > 0.0,
      supp_b=ot_prob.b > 0.0,
  )
  init_state = State(
      v1=jnp.ones(r),
      v2=jnp.ones(r),
      u1=jnp.ones(n),
      u2=jnp.ones(m),
      g=k_g,
      err=jnp.inf
  )

  state: State = fixed_point_loop.fixpoint_iter_backprop(
      cond_fn, body_fn, min_iter, max_iter, inner_iter, constants, init_state
  )

  q = state.u1[:, None] * k_q * state.v1[None, :]
  r = state.u2[:, None] * k_r * state.v2[None, :]

  return q, r, state.g


def compute_lambdas(
    const: Constants, state: State, gamma: float, g: jnp.ndarray, *,
    lse_mode: bool
) -> Tuple[float, float]:
  """TODO."""
  gamma_inv = 1.0 / gamma
  rho_a = const.rho_a
  rho_b = const.rho_b

  if lse_mode:
    num_1 = jsp.special.logsumexp((-gamma_inv / rho_a) * state.u1, b=const.a)
    num_2 = jsp.special.logsumexp((-gamma_inv / rho_b) * state.u2, b=const.b)
    den = jsp.special.logsumexp(g - (state.v1 + state.v2))
    const_1 = num_1 - den
    const_2 = num_2 - den

    ratio_1 = _get_ratio(rho_a, gamma)
    ratio_2 = _get_ratio(rho_b, gamma)
    harmonic = 1.0 / (1.0 - (ratio_1 * ratio_2))
    lam_1 = harmonic * gamma_inv * ratio_1 * (const_1 - ratio_2 * const_2)
    lam_2 = harmonic * gamma_inv * ratio_2 * (const_2 - ratio_1 * const_1)
    return lam_1, lam_2

  num_1 = jnp.sum(
      jnp.where(
          const.supp_a, ((state.u1 ** (-gamma_inv / rho_a)) * const.a), 0.0
      )
  )
  num_2 = jnp.sum(
      jnp.where(
          const.supp_b, ((state.u2 ** (-gamma_inv / rho_b)) * const.b), 0.0
      )
  )
  den = jnp.sum(g / (state.v1 * state.v2))
  const_1 = jnp.log(num_1 / den)
  const_2 = jnp.log(num_2 / den)

  ratio_1 = _get_ratio(rho_a, gamma)
  ratio_2 = _get_ratio(rho_b, gamma)
  harmonic = 1.0 / (1.0 - (ratio_1 * ratio_2))
  lam_1 = harmonic * gamma_inv * ratio_1 * (const_1 - ratio_2 * const_2)
  lam_2 = harmonic * gamma_inv * ratio_2 * (const_2 - ratio_1 * const_1)
  return lam_1, lam_2


def _rho(tau: float) -> float:
  tau = jnp.asarray(tau)  # avoid division by 0 in Python, get NaN instead
  return tau / (1.0 - tau)


def _get_ratio(rho: float, gamma: float) -> float:
  gamma_inv = 1.0 / gamma
  return jnp.where(jnp.isfinite(rho), rho / (rho + gamma_inv), 1.0)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import NamedTuple, Optional, Tuple

import jax
import jax.experimental.sparse as jesp
import jax.numpy as jnp

from ott.math import utils as mu
from ott.problems.linear import linear_problem

__all__ = [
    "UnivariateOutput",
    "uniform_solver",
    "quantile_solver",
    "north_west_solver",
]


class UnivariateOutput(NamedTuple):
  r"""Output of the univariate solver.

  Args:
    prob: Linear problem between two weighted ``[n, d]`` and ``[m, d]``
      point clouds.
    ot_costs: Array of shape ``[d,]`` of OT costs, computed independently along
      each of the :math:`d` slices.
    paired_indices: Array of shape ``[d, 2, n + m]``, of :math:`n + m` pairs
      of indices, for which the optimal transport assigns mass, on each slice
      of the :math:`d` slices described in the dataset. Namely, for each index
      :math:`0 <= k < n + m`, :math:`0 <= s < d`, if one has
      :math:`i := \text{paired_indices}[s, 0, k]` and
      :math:`j := \text{paired_indices}[s, 1, k]`, then point :math:`i` in
      the first point cloud sends mass to point :math:`j` in the second,
      in slice :math:`s`.
    mass_paired_indices: ``[d, n + m]`` array of weights. Using the notation
      above, if :math:`0 <= k < n + m`, and :math:`0 <= s < d`  then writing
      :math:`i := \text{paired_indices}[s, 0, k]` and
      :math:`j := \text{paired_indices}[s, 1, k]`, point :math:`i` sends
      :math:`\text{mass_paired_indices}[s, k]` to point :math:`j`.
    dual_a: Array of shape ``[n,]`` containing the first dual variable.
    dual_b: Array of shape ``[m,]`` containing the second dual variable.
  """
  prob: linear_problem.LinearProblem
  ot_costs: jnp.ndarray
  paired_indices: Optional[jnp.ndarray] = None
  mass_paired_indices: Optional[jnp.ndarray] = None
  dual_a: Optional[jnp.ndarray] = None
  dual_b: Optional[jnp.ndarray] = None

  @property
  def transport_matrices(self) -> jesp.BCOO:
    """Array of shape ``[d, n, m]`` containing all transport matrices.

    The matrices will be sparse, with at most :math:`n + m` entries
    for each of the :math:`d` slices.
    """
    b = len(self.ot_costs)
    n, m = self.prob.geom.shape
    data = self.mass_paired_indices
    indices = self.paired_indices.swapaxes(1, 2)
    return jesp.BCOO((data, indices), shape=(b, n, m))

  @property
  def mean_transport_matrix(self) -> jesp.BCOO:
    """Mean transport matrix, averaged over :math:`d` slices."""
    sparse_mean = jesp.sparsify(jnp.mean)
    return sparse_mean(self.transport_matrices, axis=0)

  @property
  def dual_costs(self) -> jnp.ndarray:
    """Array of shape ``[d,]`` containing the dual costs."""
    assert self.dual_a is not None, "Dual variables have not been computed."
    dual_obj = jnp.sum(self.dual_a * self.prob.a[None, :], axis=1)
    dual_obj += jnp.sum(self.dual_b * self.prob.b[None, :], axis=1)
    return dual_obj


def uniform_solver(
    prob: linear_problem.LinearProblem,
    return_transport: bool = False,
) -> UnivariateOutput:
  """Univariate solver between two equally sized and uniformly weighted distributions.

  Args:
    prob: Problem with two :class:`point clouds <ott.geometry.pointcloud.PointCloud>`
      of shapes ``[n, d]`` and ``[n, d]`` and a ground cost.
      The ``[n,]`` sized probability weights are stored
      in attributes :attr:`~ott.problems.linear.linear_problem.LinearProblem.a`
      and :attr:`~ott.problems.linear.linear_problem.LinearProblem.b`.
    return_transport: Whether to also return the mapped pairs used to compute
      the :attr:`~ott.solvers.linear.univariate.UnivariateOutput.transport_matrices`.

  Returns:
    The univariate output. Note that the
    :attr:`~ott.solvers.linear.univariate.UnivariateOutput.mass_paired_indices`
    can be :math:`0` in some entries, but always sums to :math:`1`
    for each of the :math:`d` slices.
  """  # noqa: E501

  @functools.partial(jax.vmap, in_axes=[1, 1])
  @functools.partial(jax.vmap, in_axes=[0, 0])
  def cost(x: jnp.ndarray, y: jnp.ndarray) -> float:
    return cost_fn(x[None], y[None])

  assert prob.is_equal_size, "Source and target have different sizes."
  assert prob.is_uniform, "Source or target marginals are not uniform."

  geom = prob.geom
  n, _ = geom.shape
  x, y, cost_fn = geom.x, geom.y, geom.cost_fn

  i_x, i_y = jnp.argsort(x, axis=0), jnp.argsort(y, axis=0)
  x = jnp.take_along_axis(x, i_x, axis=0)
  y = jnp.take_along_axis(y, i_y, axis=0)
  ot_costs = jnp.mean(cost(x, y), axis=-1)  # (d, n) -> (d,)

  if return_transport:
    paired_indices = jnp.stack([i_x, i_y]).transpose([2, 0, 1])
    mass_paired_indices = jnp.ones((len(ot_costs), n)) / n
  else:
    paired_indices = mass_paired_indices = None

  return UnivariateOutput(
      prob,
      ot_costs,
      paired_indices=paired_indices,
      mass_paired_indices=mass_paired_indices,
  )


def quantile_solver(
    prob: linear_problem.LinearProblem,
    return_transport: bool = False,
) -> UnivariateOutput:
  """Univariate solver between quantile functions of distributions.

  Args:
    prob: Problem with two :class:`point clouds <ott.geometry.pointcloud.PointCloud>`
      of shapes ``[n, d]`` and ``[m, d]`` and a ground cost.
      The ``[n,]`` and ``[m,]`` sized probability weights vectors are stored
      in attributes :attr:`~ott.problems.linear.linear_problem.LinearProblem.a`
      and :attr:`~ott.problems.linear.linear_problem.LinearProblem.b`.
    return_transport: Whether to also return the mapped pairs used to compute
      the :attr:`~ott.solvers.linear.univariate.UnivariateOutput.transport_matrices`.

  Returns:
    The univariate output. Note that the
    :attr:`~ott.solvers.linear.univariate.UnivariateOutput.mass_paired_indices`
    can be :math:`0` in some entries, but always sums to :math:`1`
    for each of the :math:`d` slices.

  Notes:
    This function was inspired by :func:`~scipy.stats.wasserstein_distance`,
    but can be used with other costs, not just :math:`c(x, y) = |x - y|`.
  """  # noqa: E501

  @functools.partial(jax.vmap, in_axes=[1, 1])
  def dist(x: jnp.ndarray, y: jnp.ndarray):
    x, i_x = mu.sort_and_argsort(x, argsort=True)
    y, i_y = mu.sort_and_argsort(y, argsort=True)

    all_values = jnp.concatenate([x, y])
    all_values_sorted, all_values_sorter = mu.sort_and_argsort(
        all_values, argsort=True
    )

    x_pdf = jnp.concatenate([prob.a[i_x], jnp.zeros_like(prob.b)])
    x_pdf = x_pdf[all_values_sorter]
    y_pdf = jnp.concatenate([jnp.zeros_like(prob.a), prob.b[i_y]])
    y_pdf = y_pdf[all_values_sorter]

    x_cdf = jnp.cumsum(x_pdf)
    y_cdf = jnp.cumsum(y_pdf)

    x_y_cdfs = jnp.concatenate([x_cdf, y_cdf])
    quantile_levels, _ = mu.sort_and_argsort(x_y_cdfs, argsort=False)

    i_x_cdf_inv = jnp.searchsorted(x_cdf, quantile_levels)
    x_cdf_inv = all_values_sorted[i_x_cdf_inv]
    i_y_cdf_inv = jnp.searchsorted(y_cdf, quantile_levels)
    y_cdf_inv = all_values_sorted[i_y_cdf_inv]

    diff_q = jnp.diff(quantile_levels)
    successive_costs = jax.vmap(
        prob.geom.cost_fn, in_axes=[0, 0]
    )(x_cdf_inv[1:, None], y_cdf_inv[1:, None])
    cost = jnp.sum(successive_costs * diff_q)

    if not return_transport:
      return cost, None, None

    n = x.shape[0]
    i_in_sorted_x_of_quantile = all_values_sorter[i_x_cdf_inv] % n
    i_in_sorted_y_of_quantile = all_values_sorter[i_y_cdf_inv] - n
    orig_i = i_x[i_in_sorted_x_of_quantile][1:]
    orig_j = i_y[i_in_sorted_y_of_quantile][1:]
    paired_indices, mass_paired_indices = jnp.stack([orig_i, orig_j]), diff_q

    return cost, paired_indices, mass_paired_indices

  ot_costs, pi, mpi = dist(prob.geom.x, prob.geom.y)
  return UnivariateOutput(
      prob,
      ot_costs,
      paired_indices=pi,
      mass_paired_indices=mpi,
  )


def north_west_solver(prob: linear_problem.LinearProblem) -> UnivariateOutput:
  """Univariate solver that implements the north-west corner rule.

  This rule is described in :cite:`peyre:19`, sec. 3.4.2 and the dual variables
  are stored as described in :cite:`sejourne:22`, alg. 3.

  Args:
    prob: Problem with two :class:`point clouds <ott.geometry.pointcloud.PointCloud>`
      of shapes ``[n, d]`` and ``[m, d]`` and a ground cost.
      The ``[n,]`` and ``[m,]`` sized probability weights are stored
      in attributes :attr:`~ott.problems.linear.linear_problem.LinearProblem.a`
      and :attr:`~ott.problems.linear.linear_problem.LinearProblem.b`.

  Returns:
    The univariate output. Note that the
    :attr:`~ott.solvers.linear.univariate.UnivariateOutput.mass_paired_indices`
    can be :math:`0` in some entries, but always sums to :math:`1`
    for each of the :math:`d` slices.
  """  # noqa: E501

  class State(NamedTuple):
    x: jnp.ndarray
    y: jnp.ndarray
    a: jnp.ndarray
    b: jnp.ndarray
    paired_indices: jnp.ndarray
    mass_paired_indices: jnp.ndarray
    dual_a: jnp.ndarray
    dual_b: jnp.ndarray

  def dual_a_update(state: State, i: int,
                    j: int) -> Tuple[State, jnp.ndarray, jnp.ndarray]:
    next_ixs = jnp.array([i + 1, j])
    val = cost_fn(state.x[i + 1, None], state.y[j, None]) - state.dual_b[j]
    da = state.dual_a.at[i + 1].set(val)
    return state._replace(dual_a=da), state.a[i], next_ixs

  def dual_b_update(state: State, i: int,
                    j: int) -> Tuple[State, jnp.ndarray, jnp.ndarray]:
    next_ixs = jnp.array([i, j + 1])
    val = cost_fn(state.x[i, None], state.y[j + 1, None]) - state.dual_a[i]
    db = state.dual_b.at[j + 1].set(val)
    return state._replace(dual_b=db), state.b[j], next_ixs

  def body_fun(ix: int, state: State) -> State:
    i, j = state.paired_indices[0, ix], state.paired_indices[1, ix]
    state, min_ab, next_ixs = jax.lax.cond(
        state.a[i] < state.b[j], dual_a_update, dual_b_update, state, i, j
    )

    pi = state.paired_indices.at[:, ix + 1].set(next_ixs)
    mpi = state.mass_paired_indices.at[ix].set(min_ab)
    a_ = state.a.at[i].set(state.a[i] - min_ab)
    b_ = state.b.at[j].set(state.b[j] - min_ab)

    return state._replace(
        paired_indices=pi, mass_paired_indices=mpi, a=a_, b=b_
    )

  @functools.partial(jax.vmap, in_axes=[1, 1])
  def dist(x: jnp.ndarray, y: jnp.ndarray):
    x, i_x = mu.sort_and_argsort(x, argsort=True)
    y, i_y = mu.sort_and_argsort(y, argsort=True)
    sorted_a, sorted_b = a[i_x], b[i_y]

    paired_indices = jnp.zeros((2, q), dtype=int)
    mass_paired_indices = jnp.zeros(q)

    state = State(
        x,
        y,
        a=sorted_a,
        b=sorted_b,
        paired_indices=paired_indices,
        mass_paired_indices=mass_paired_indices,
        dual_a=jnp.zeros(n),
        dual_b=jnp.zeros(m).at[0].set(cost_fn(x[0, None], y[0, None])),
    )
    state = jax.lax.fori_loop(0, q - 1, body_fun, state)

    ot_cost = jnp.sum(state.dual_a * sorted_a
                     ) + jnp.sum(state.dual_b * sorted_b)

    p_final = jnp.maximum(state.a[-1], state.b[-1])
    mass_paired_indices = state.mass_paired_indices.at[-1].set(p_final)

    return (
        ot_cost,
        state.paired_indices,
        mass_paired_indices,
        # restore the original order
        state.dual_a[jnp.argsort(i_x)],
        state.dual_b[jnp.argsort(i_y)],
    )

  a, b = prob.a, prob.b
  n, m = prob.geom.shape
  q = m + n - 1
  cost_fn = prob.geom.cost_fn

  ot_costs, pi, mpi, dual_a, dual_b = dist(prob.geom.x, prob.geom.y)
  return UnivariateOutput(
      prob,
      ot_costs,
      paired_indices=pi,
      mass_paired_indices=mpi,
      dual_a=dual_a,
      dual_b=dual_b,
  )



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Dict, Literal, Optional, Sequence, Tuple

import jax
import jax.numpy as jnp
import jax.scipy as jsp
import jax.tree_util as jtu
import numpy as np

from ott.geometry import costs
from ott.problems.linear import linear_problem

try:
  import matplotlib as mpl
  import matplotlib.pyplot as plt
except ImportError:
  mpl = plt = None

__all__ = ["DualPotentials", "EntropicPotentials"]
Potential_t = Callable[[jnp.ndarray], float]


@jtu.register_pytree_node_class
class DualPotentials:
  r"""The Kantorovich dual potential functions :math:`f` and :math:`g`.

  :math:`f` and :math:`g` are a pair of functions, candidates for the dual
  OT Kantorovich problem, supposedly optimal for a given pair of measures.

  Args:
    f: The first dual potential function.
    g: The second dual potential function.
    cost_fn: The cost function used to solve the OT problem.
    corr: Whether the duals solve the problem in distance form, or correlation
      form (as used for instance for ICNNs, see, e.g., top right of p.3 in
      :cite:`makkuva:20`)
  """

  def __init__(
      self,
      f: Potential_t,
      g: Potential_t,
      *,
      cost_fn: costs.CostFn,
      corr: bool = False
  ):
    self._f = f
    self._g = g
    assert (
        not corr or type(cost_fn) is costs.SqEuclidean
    ), "Duals in `corr` form can only be used with a squared-Euclidean cost."
    self.cost_fn = cost_fn
    self._corr = corr

  def transport(self, vec: jnp.ndarray, forward: bool = True) -> jnp.ndarray:
    r"""Transport ``vec`` according to Gangbo-McCann Brenier :cite:`brenier:91`.

    Uses Proposition 1.15 from :cite:`santambrogio:15` to compute an OT map when
    applying the inverse gradient of cost.

    When the cost is a general cost, the operator uses the
    :meth:`~ott.geometry.costs.CostFn.twist_operator` associated of the
    corresponding :class:`~ott.geometry.costs.CostFn`.

    When the cost is a translation invariant :class:`~ott.geometry.costs.TICost`
    cost, :math:`c(x,y)=h(x-y)`, and the twist operator translates to the
    application of the convex conjugate of :math:`h` to the
    gradient of the dual potentials, namely
    :math:`x- (\nabla h^*)\circ \nabla f(x)` for the forward map,
    where :math:`h^*` is the Legendre transform of :math:`h`. For instance,
    in the case :math:`h(\cdot) = \|\cdot\|^2, \nabla h(\cdot) = 2 \cdot\,`,
    one has :math:`h^*(\cdot) = \|.\|^2 / 4`, and therefore
    :math:`\nabla h^*(\cdot) = 0.5 \cdot\,`.

    Note:
      When the dual potentials are solved in correlation form, and marked
      accordingly by setting ``corr`` to ``True``, the maps are
      :math:`\nabla g` for forward, :math:`\nabla f` for backward map. This can
      only make sense when using the squared-Euclidean
      :class:`~ott.geometry.costs.SqEuclidean` cost.

    Args:
      vec: Points to transport, array of shape ``[n, d]``.
      forward: Whether to transport the points from source to the target
        distribution or vice-versa.

    Returns:
      The transported points.
    """
    from ott.geometry import costs

    vec = jnp.atleast_2d(vec)

    if self._corr and isinstance(self.cost_fn, costs.SqEuclidean):
      return self._grad_f(vec) if forward else self._grad_g(vec)
    twist_op = jax.vmap(self.cost_fn.twist_operator, in_axes=[0, 0, None])
    if forward:
      return twist_op(vec, self._grad_f(vec), False)
    return twist_op(vec, self._grad_g(vec), True)

  def distance(self, src: jnp.ndarray, tgt: jnp.ndarray) -> float:
    r"""Evaluate Wasserstein distance between samples using dual potentials.

    This uses direct estimation of potentials against measures when dual
    functions are provided in usual form. This expression is valid for any
    cost function.

    When potentials are given in correlation form, as specified by the flag
    ``corr``, the dual potentials solve the dual problem corresponding to the
    minimization of the primal OT problem where the ground cost is
    :math:`-2\langle x,y\rangle`. To recover the (squared) 2-Wasserstein
    distance, terms are re-arranged and contributions from squared norms are
    taken into account.

    Args:
      src: Samples from the source distribution, array of shape ``[n, d]``.
      tgt: Samples from the target distribution, array of shape ``[m, d]``.

    Returns:
      Wasserstein distance using specified cost function.
    """
    src, tgt = jnp.atleast_2d(src), jnp.atleast_2d(tgt)
    f = jax.vmap(self.f)
    g = jax.vmap(self.g)
    out = jnp.mean(f(src)) + jnp.mean(g(tgt))
    if self._corr:
      out = -2.0 * out + jnp.mean(jnp.sum(src ** 2, axis=-1))
      out += jnp.mean(jnp.sum(tgt ** 2, axis=-1))
    return out

  @property
  def f(self) -> Potential_t:
    """The first dual potential function."""
    return self._f

  @property
  def g(self) -> Potential_t:
    """The second dual potential function."""
    return self._g

  @property
  def _grad_f(self) -> Callable[[jnp.ndarray], jnp.ndarray]:
    """Vectorized gradient of the potential function :attr:`f`."""
    return jax.vmap(jax.grad(self.f, argnums=0))

  @property
  def _grad_g(self) -> Callable[[jnp.ndarray], jnp.ndarray]:
    """Vectorized gradient of the potential function :attr:`g`."""
    return jax.vmap(jax.grad(self.g, argnums=0))

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return [], {
        "f": self._f,
        "g": self._g,
        "cost_fn": self.cost_fn,
        "corr": self._corr
    }

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "DualPotentials":
    return cls(*children, **aux_data)

  def plot_ot_map(
      self,
      source: jnp.ndarray,
      target: jnp.ndarray,
      samples: Optional[jnp.ndarray] = None,
      forward: bool = True,
      ax: Optional["plt.Axes"] = None,
      scatter_kwargs: Optional[Dict[str, Any]] = None,
      legend_kwargs: Optional[Dict[str, Any]] = None,
  ) -> Tuple["plt.Figure", "plt.Axes"]:
    """Plot data and learned optimal transport map.

    Args:
      source: samples from the source measure
      target: samples from the target measure
      samples: extra samples to transport, either ``source`` (if ``forward``) or
        ``target`` (if not ``forward``) by default.
      forward: use the forward map from the potentials if ``True``,
        otherwise use the inverse map.
      ax: axis to add the plot to
      scatter_kwargs: additional kwargs passed into
        :meth:`~matplotlib.axes.Axes.scatter`
      legend_kwargs: additional kwargs passed into
        :meth:`~matplotlib.axes.Axes.legend`

    Returns:
      Figure and axes.
    """
    if mpl is None:
      raise RuntimeError("Please install `matplotlib` first.")

    if scatter_kwargs is None:
      scatter_kwargs = {"alpha": 0.5}
    if legend_kwargs is None:
      legend_kwargs = {
          "ncol": 3,
          "loc": "upper center",
          "bbox_to_anchor": (0.5, -0.05),
          "edgecolor": "k"
      }

    if ax is None:
      fig = plt.figure(facecolor="white")
      ax = fig.add_subplot(111)
    else:
      fig = ax.get_figure()

    # plot the source and target samples
    if forward:
      label_transport = r"$\nabla f(source)$"
      source_color, target_color = "#1A254B", "#A7BED3"
    else:
      label_transport = r"$\nabla g(target)$"
      source_color, target_color = "#A7BED3", "#1A254B"

    ax.scatter(
        source[:, 0],
        source[:, 1],
        color=source_color,
        label="source",
        **scatter_kwargs,
    )
    ax.scatter(
        target[:, 0],
        target[:, 1],
        color=target_color,
        label="target",
        **scatter_kwargs,
    )

    # plot the transported samples
    samples = (source if forward else target) if samples is None else samples
    transported_samples = self.transport(samples, forward=forward)
    ax.scatter(
        transported_samples[:, 0],
        transported_samples[:, 1],
        color="#F2545B",
        label=label_transport,
        **scatter_kwargs,
    )

    for i in range(samples.shape[0]):
      ax.arrow(
          samples[i, 0],
          samples[i, 1],
          transported_samples[i, 0] - samples[i, 0],
          transported_samples[i, 1] - samples[i, 1],
          color=[0.5, 0.5, 1],
          alpha=0.3,
      )

    ax.legend(**legend_kwargs)
    return fig, ax

  def plot_potential(
      self,
      forward: bool = True,
      quantile: float = 0.05,
      kantorovich: bool = True,
      ax: Optional["mpl.axes.Axes"] = None,
      x_bounds: Tuple[float, float] = (-6, 6),
      y_bounds: Tuple[float, float] = (-6, 6),
      num_grid: int = 50,
      contourf_kwargs: Optional[Dict[str, Any]] = None,
  ) -> Tuple["mpl.figure.Figure", "mpl.axes.Axes"]:
    r"""Plot the potential.

    Args:
      forward: use the forward map from the potentials
        if ``True``, otherwise use the inverse map
      quantile: quantile to filter the potentials with
      kantorovich: whether to plot the Kantorovich potential
      ax: axis to add the plot to
      x_bounds: x-axis bounds of the plot
        :math:`(x_{\text{min}}, x_{\text{max}})`
      y_bounds: y-axis bounds of the plot
        :math:`(y_{\text{min}}, y_{\text{max}})`
      num_grid: number of points to discretize the domain into a grid
        along each dimension
      contourf_kwargs: additional kwargs passed into
        :meth:`~matplotlib.axes.Axes.contourf`

    Returns:
      Figure and axes.
    """
    if contourf_kwargs is None:
      contourf_kwargs = {}

    ax_specified = ax is not None
    if not ax_specified:
      fig, ax = plt.subplots(figsize=(6, 6), facecolor="white")
    else:
      fig = ax.get_figure()

    x1 = jnp.linspace(*x_bounds, num=num_grid)
    x2 = jnp.linspace(*y_bounds, num=num_grid)
    X1, X2 = jnp.meshgrid(x1, x2)
    X12flat = jnp.hstack((X1.reshape(-1, 1), X2.reshape(-1, 1)))
    Zflat = jax.vmap(self.f if forward else self.g)(X12flat)
    if kantorovich:
      Zflat = 0.5 * (jnp.linalg.norm(X12flat, axis=-1) ** 2) - Zflat
    Zflat = np.asarray(Zflat)
    vmin, vmax = np.quantile(Zflat, [quantile, 1.0 - quantile])
    Zflat = Zflat.clip(vmin, vmax)
    Z = Zflat.reshape(X1.shape)

    CS = ax.contourf(X1, X2, Z, cmap="Blues", **contourf_kwargs)
    ax.set_xlim(*x_bounds)
    ax.set_ylim(*y_bounds)
    fig.colorbar(CS, ax=ax)
    if not ax_specified:
      fig.tight_layout()
    ax.set_title(r"$f$" if forward else r"$g$")
    return fig, ax


@jtu.register_pytree_node_class
class EntropicPotentials(DualPotentials):
  """Dual potential functions from finite samples :cite:`pooladian:21`.

  Args:
    f_xy: The first dual potential vector of shape ``[n,]``.
    g_xy: The second dual potential vector of shape ``[m,]``.
    prob: Linear problem with :class:`~ott.geometry.pointcloud.PointCloud`
      geometry that was used to compute the dual potentials using, e.g.,
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn`.
    f_xx: The first dual potential vector of shape ``[n,]`` used for debiasing
      :cite:`pooladian:22`.
    g_yy: The second dual potential vector of shape ``[m,]`` used for debiasing.
  """

  def __init__(
      self,
      f_xy: jnp.ndarray,
      g_xy: jnp.ndarray,
      prob: linear_problem.LinearProblem,
      f_xx: Optional[jnp.ndarray] = None,
      g_yy: Optional[jnp.ndarray] = None,
  ):
    # we pass directly the arrays and override the properties
    # since only the properties need to be callable
    super().__init__(f_xy, g_xy, cost_fn=prob.geom.cost_fn, corr=False)
    self._prob = prob
    self._f_xx = f_xx
    self._g_yy = g_yy

  @property
  def f(self) -> Potential_t:  # noqa: D102
    return self._potential_fn(kind="f")

  @property
  def g(self) -> Potential_t:  # noqa: D102
    return self._potential_fn(kind="g")

  def _potential_fn(self, *, kind: Literal["f", "g"]) -> Potential_t:
    from ott.geometry import pointcloud

    def callback(
        x: jnp.ndarray,
        *,
        potential: jnp.ndarray,
        y: jnp.ndarray,
        weights: jnp.ndarray,
        epsilon: float,
    ) -> float:
      x = jnp.atleast_2d(x)
      assert x.shape[-1] == y.shape[-1], (x.shape, y.shape)
      geom = pointcloud.PointCloud(x, y, cost_fn=self.cost_fn)
      cost = geom.cost_matrix
      z = (potential - cost) / epsilon
      lse = -epsilon * jsp.special.logsumexp(z, b=weights, axis=-1)
      return jnp.squeeze(lse)

    assert isinstance(
        self._prob.geom, pointcloud.PointCloud
    ), f"Expected point cloud geometry, found `{type(self._prob.geom)}`."
    x, y = self._prob.geom.x, self._prob.geom.y
    a, b = self._prob.a, self._prob.b

    # `f_xx` or `g_yy` can both be `None`, we check for this later
    debiased_potentials = EntropicPotentials(self._f_xx, self._g_yy, self._prob)
    if kind == "f":
      # When seeking to evaluate 1st potential function,
      # the 2nd set of potential values and support should be used,
      # see proof of Prop. 2 in https://arxiv.org/pdf/2109.12004.pdf
      potential, arr, weights = self._g, y, b
      potential_other = None if self._f_xx is None else debiased_potentials.g
    else:
      potential, arr, weights = self._f, x, a
      potential_other = None if self._g_yy is None else debiased_potentials.f

    potential_xy = jax.tree_util.Partial(
        callback,
        potential=potential,
        y=arr,
        weights=weights,
        epsilon=self.epsilon,
    )

    if potential_other is None:
      return potential_xy
    return lambda x: (potential_xy(x) - potential_other(x))

  @property
  def is_debiased(self) -> bool:
    """Whether the :attr:`f` or :attr:`g` is debiased.

    The :attr:`g` potential is **not** debiased when ``static_b = True`` is
    passed in :func:`~ott.tools.sinkhorn_divergence.sinkhorn_divergence`.
    """
    return self._f_xx is not None or self._g_yy is not None

  @property
  def epsilon(self) -> float:
    """Entropy regularizer."""
    return self._prob.geom.epsilon

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return [self._f, self._g, self._prob, self._f_xx, self._g_yy], {}



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
from typing import Any, Callable, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp

import optax
from flax import linen as nn
from flax import struct
from flax.core import frozen_dict
from flax.training import train_state

__all__ = ["PotentialTrainState", "BasePotential", "PotentialMLP", "MLP"]

PotentialValueFn_t = Callable[[jnp.ndarray], jnp.ndarray]
PotentialGradientFn_t = Callable[[jnp.ndarray], jnp.ndarray]


class PotentialTrainState(train_state.TrainState):
  """Adds information about the model's value and gradient to the state.

  This extends :class:`~flax.training.train_state.TrainState` to include
  the potential methods from the
  :class:`~ott.neural.networks.potentials.BasePotential` used during training.

  Args:
    potential_value_fn: the potential's value function
    potential_gradient_fn: the potential's gradient function
  """
  potential_value_fn: Callable[
      [frozen_dict.FrozenDict[str, jnp.ndarray], Optional[PotentialValueFn_t]],
      PotentialValueFn_t] = struct.field(pytree_node=False)
  potential_gradient_fn: Callable[[frozen_dict.FrozenDict[str, jnp.ndarray]],
                                  PotentialGradientFn_t] = struct.field(
                                      pytree_node=False
                                  )


class BasePotential(abc.ABC, nn.Module):
  """Base class for the neural solver models."""

  @property
  @abc.abstractmethod
  def is_potential(self) -> bool:
    """Indicates if the module implements a potential value or a vector field.

    Returns:
      ``True`` if the module defines a potential, ``False`` if it defines a
       vector field.
    """

  def potential_value_fn(
      self,
      params: frozen_dict.FrozenDict[str, jnp.ndarray],
      other_potential_value_fn: Optional[PotentialValueFn_t] = None,
  ) -> PotentialValueFn_t:
    r"""Return a function giving the value of the potential.

    Applies the module if :attr:`is_potential` is ``True``, otherwise
    constructs the value of the potential from the gradient with

    .. math::

      g(y) = -f(\nabla_y g(y)) + y^T \nabla_y g(y)

    where :math:`\nabla_y g(y)` is detached for the envelope theorem
    :cite:`danskin:67,bertsekas:71`
    to give the appropriate first derivatives of this construction.

    Args:
      params: parameters of the module
      other_potential_value_fn: function giving the value of the other
        potential. Only needed when :attr:`is_potential` is ``False``.

    Returns:
      A function that can be evaluated to obtain a potential value, or a linear
      interpolation of a potential.
    """
    if self.is_potential:
      return lambda x: self.apply({"params": params}, x)

    assert other_potential_value_fn is not None, \
      "The value of the gradient-based potential depends " \
      "on the value of the other potential."

    def value_fn(x: jnp.ndarray) -> jnp.ndarray:
      squeeze = x.ndim == 1
      if squeeze:
        x = jnp.expand_dims(x, 0)
      grad_g_x = jax.lax.stop_gradient(self.apply({"params": params}, x))
      value = -other_potential_value_fn(grad_g_x) + \
              jax.vmap(jnp.dot)(grad_g_x, x)
      return value.squeeze(0) if squeeze else value

    return value_fn

  def potential_gradient_fn(
      self,
      params: frozen_dict.FrozenDict[str, jnp.ndarray],
  ) -> PotentialGradientFn_t:
    """Return a function returning a vector or the gradient of the potential.

    Args:
      params: parameters of the module

    Returns:
      A function that can be evaluated to obtain the potential's gradient
    """
    if self.is_potential:
      return jax.vmap(jax.grad(self.potential_value_fn(params)))
    return lambda x: self.apply({"params": params}, x)

  def create_train_state(
      self,
      rng: jax.Array,
      optimizer: optax.OptState,
      input: Union[int, Tuple[int, ...]],
      **kwargs: Any,
  ) -> PotentialTrainState:
    """Create initial training state."""
    params = self.init(rng, jnp.ones(input))["params"]

    return PotentialTrainState.create(
        apply_fn=self.apply,
        params=params,
        tx=optimizer,
        potential_value_fn=self.potential_value_fn,
        potential_gradient_fn=self.potential_gradient_fn,
        **kwargs,
    )


class PotentialMLP(BasePotential):
  """Potential MLP.

  Args:
    dim_hidden: Sequence specifying the size of hidden dimensions. The output
      dimension of the last layer is automatically set to 1 if
      :attr:`is_potential` is ``True``, or the dimension of the input otherwise.
    is_potential: Model the potential if ``True``, otherwise
      model the gradient of the potential.
    act_fn: Activation function.
  """

  dim_hidden: Sequence[int]
  is_potential: bool = True
  act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.leaky_relu

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> jnp.ndarray:  # noqa: D102
    squeeze = x.ndim == 1
    if squeeze:
      x = jnp.expand_dims(x, 0)
    assert x.ndim == 2, x.ndim
    n_input = x.shape[-1]

    z = x
    for n_hidden in self.dim_hidden:
      Wx = nn.Dense(n_hidden, use_bias=True)
      z = self.act_fn(Wx(z))

    if self.is_potential:
      Wx = nn.Dense(1, use_bias=True)
      z = Wx(z).squeeze(-1)

      quad_term = 0.5 * jax.vmap(jnp.dot)(x, x)
      z += quad_term
    else:
      Wx = nn.Dense(n_input, use_bias=True)
      z = x + Wx(z)

    return z.squeeze(0) if squeeze else z


class MLP(nn.Module):
  """A simple MLP model of a potential used in default initialization.

  Args:
    dim_hidden: Sequence specifying the size of hidden dimensions.
    act_fn: Activation function.
  """

  dim_hidden: Sequence[int]
  act_fn: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.elu

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
    """Apply MLP transform."""
    for feat in self.dim_hidden[:-1]:
      x = self.act_fn(nn.Dense(feat)(x))
    return nn.Dense(self.dim_hidden[-1])(x)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib

from . import (
    datasets,
    experimental,
    geometry,
    initializers,
    math,
    problems,
    solvers,
    tools,
    utils,
)

with contextlib.suppress(ImportError):
  from . import neural

from ._version import __version__

del contextlib



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from types import MappingProxyType
from typing import Any, Mapping, Optional, Tuple, Type, Union

import jax
import jax.numpy as jnp

from ott import utils
from ott.geometry import costs, geometry, pointcloud, segment
from ott.problems.linear import linear_problem, potentials
from ott.solvers import linear
from ott.solvers.linear import acceleration, sinkhorn, sinkhorn_lr

__all__ = [
    "sinkhorn_divergence", "segment_sinkhorn_divergence",
    "SinkhornDivergenceOutput"
]

Potentials = Tuple[jnp.ndarray, jnp.ndarray]
Factors = Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]


@utils.register_pytree_node
class SinkhornDivergenceOutput:  # noqa: D101
  r"""Holds the outputs of a call to :func:`sinkhorn_divergence`.

  Objects of this class contain both solutions and problem definition of a
  two or three regularized OT problem instantiated when computing a Sinkhorn
  divergence between two probability distributions.

  Args:
    divergence: value of the Sinkhorn divergence
    geoms: three geometries describing the Sinkhorn divergence, of respective
      sizes ``[n, m], [n, n], [m, m]`` if their cost or kernel matrices where
      instantiated.
    a: first ``[n,]`` vector of marginal weights.
    b: second ``[m,]`` vector of marginal weights.
    potentials: three pairs of dual potential vectors, of sizes
      ``[n,], [m,]``, ``[n,], [n,]``, ``[m,], [m,]``, returned when the call
      to the :func:`~ott.solvers.linear.solve` solver to compute the divergence
      relies on a vanilla :class:`~ott.solver.linear.sinkhorn.Sinkhorn` solver.
    factors: three triplets of matrices, of sizes
      ``([n, rank], [m, rank], [rank,])``, ``([n, rank], [n, rank], [rank,])``
      and ``([m, rank], [m, rank], [rank,])``, returned when the call
      to the :func:`~ott.solvers.linear.solve` solver to compute the divergence
      relies on a low-rank :class:`~ott.solver.linear.sinkhorn_lr.LRSinkhorn`
      solver.
    converged: triplet of booleans indicating the convergence of each of the
      three problems run to compute the divergence.
    n_iters: number of iterations keeping track of compute effort needed to
      complete each of the three terms in the divergence.
  """
  divergence: float
  geoms: Tuple[geometry.Geometry, geometry.Geometry, geometry.Geometry]
  a: jnp.ndarray
  b: jnp.ndarray
  potentials: Optional[Tuple[Potentials, Potentials, Potentials]]
  factors: Optional[Tuple[Factors, Factors, Factors]]
  errors: Tuple[Optional[jnp.ndarray], Optional[jnp.ndarray],
                Optional[jnp.ndarray]]
  converged: Tuple[bool, bool, bool]
  n_iters: Tuple[int, int, int]

  def to_dual_potentials(self) -> "potentials.EntropicPotentials":
    """Return dual potential functions, :cite:`pooladian:22`.

    Using vectors stored in ``potentials``, instantiate a
    :class:`~ott.problems.linear.potentials.EntropicPotentials` object that will
    provide approximations to optimal dual potential functions for the dual
    OT problem defined for the geometry stored in ``geoms[0]``. These correspond
    to Equation 8 in :cite:`pooladian:22`.
    """
    assert not self.is_low_rank, \
      "Dual potentials not available: divergence computed with low-rank solver."
    geom_xy, *_ = self.geoms
    prob_xy = linear_problem.LinearProblem(geom_xy, a=self.a, b=self.b)
    (f_xy, g_xy), (f_x, _), (_, g_y) = self.potentials
    return potentials.EntropicPotentials(
        f_xy, g_xy, prob_xy, f_xx=f_x, g_yy=g_y
    )

  @property
  def is_low_rank(self) -> bool:
    """Whether the output is low-rank."""
    return self.factors is not None

  def tree_flatten(self):  # noqa: D102
    return [
        self.divergence, self.geoms, self.a, self.b, self.potentials,
        self.factors
    ], {
        "errors": self.errors,
        "n_iters": self.n_iters,
        "converged": self.converged,
    }

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    return cls(*children, **aux_data)


def sinkhorn_divergence(
    geom: Type[geometry.Geometry],
    *args: Any,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    solve_kwargs: Mapping[str, Any] = MappingProxyType({}),
    static_b: bool = False,
    share_epsilon: bool = True,
    symmetric_sinkhorn: bool = True,
    **kwargs: Any,
) -> Tuple[jnp.ndarray, SinkhornDivergenceOutput]:
  """Compute Sinkhorn divergence defined by a geometry, weights, parameters.

  Args:
    geom: Type of the geometry.
    args: Positional arguments to
      :meth:`~ott.geometry.geometry.Geometry.prepare_divergences` that are
      specific to each geometry.
    a: the weight of each input point. The sum of all elements of `a` must
      match that of `b` to converge.
    b: the weight of each target point. The sum of all elements of `b` must
      match that of `a` to converge.
    solve_kwargs: keywords arguments for
      :func:`~ott.solvers.linear.solve` that is called either twice
      if ``static_b == True`` or three times when ``static_b == False``.
    static_b: if ``True``, divergence of measure `b` against itself is **not**
      computed.
    share_epsilon: if True, enforces that the same epsilon regularizer is shared
      for all 2 or 3 terms of the Sinkhorn divergence. In that case, the epsilon
      will be by default that used when comparing x to y (contained in the first
      geometry). This flag is set to True by default, because in the default
      setting, the epsilon regularization is a function of the std of the
      entries in the cost matrix.
    symmetric_sinkhorn: Use Sinkhorn updates in Eq. 25 of :cite:`feydy:19` for
      symmetric terms comparing x/x and y/y.
    kwargs: keywords arguments to the generic class. This is specific to each
      geometry.

  Returns:
    The Sinkhorn divergence value, and output object detailing computations.
  """
  geoms = geom.prepare_divergences(*args, static_b=static_b, **kwargs)
  geom_xy, geom_x, geom_y, *_ = geoms + (None,) * 3
  num_a, num_b = geom_xy.shape

  if share_epsilon:
    if isinstance(geom_x, geometry.Geometry):
      geom_x = geom_x.copy_epsilon(geom_xy)
    if isinstance(geom_y, geometry.Geometry):
      geom_y = geom_y.copy_epsilon(geom_xy)

  a = jnp.ones(num_a) / num_a if a is None else a
  b = jnp.ones(num_b) / num_b if b is None else b
  out = _sinkhorn_divergence(
      geom_xy,
      geom_x,
      geom_y,
      a=a,
      b=b,
      symmetric_sinkhorn=symmetric_sinkhorn,
      **solve_kwargs
  )
  return out.divergence, out


def _sinkhorn_divergence(
    geometry_xy: geometry.Geometry,
    geometry_xx: geometry.Geometry,
    geometry_yy: Optional[geometry.Geometry],
    a: jnp.ndarray,
    b: jnp.ndarray,
    symmetric_sinkhorn: bool,
    **kwargs: Any,
) -> SinkhornDivergenceOutput:
  """Compute the (unbalanced) Sinkhorn divergence for the wrapper function.

    This definition includes a correction depending on the total masses of each
    measure, as defined in :cite:`sejourne:19`, eq. 15, and is extended to also
    accept :class:`~ott.solvers.linear.sinkhorn_lr.LRSinkhorn` solvers, as
    advocated in :cite:`scetbon:22b`.

  Args:
    geometry_xy: a Cost object able to apply kernels with a certain epsilon,
    between the views X and Y.
    geometry_xx: a Cost object able to apply kernels with a certain epsilon,
    between elements of the view X.
    geometry_yy: a Cost object able to apply kernels with a certain epsilon,
    between elements of the view Y.
    a: jnp.ndarray<float>[n]: the weight of each input point. The sum of
     all elements of ``b`` must match that of ``a`` to converge.
    b: jnp.ndarray<float>[m]: the weight of each target point. The sum of
     all elements of ``b`` must match that of ``a`` to converge.
    symmetric_sinkhorn: Use Sinkhorn updates in Eq. 25 of :cite:`feydy:19` for
      symmetric terms comparing x/x and y/y.
    kwargs: Keyword arguments to :func:`~ott.solvers.linear.solve`.

  Returns:
    The output object
  """
  kwargs_symmetric = kwargs.copy()
  is_low_rank = kwargs.get("rank", -1) > 0

  if symmetric_sinkhorn and not is_low_rank:
    # When computing a Sinkhorn divergence, the (x,y) terms and (x,x) / (y,y)
    # terms are computed independently. The user might want to pass some
    # kwargs to parameterize the solver's behavior, but those should
    # only apply to the (x,y) part.
    #
    # When using the Sinkhorn solver, for the (x,x) / (y,y) part, we fall back
    # on a simpler choice (parallel_dual_updates + momentum 0.5) that is known
    # to work well in such settings.
    #
    # Since symmetric terms are computed assuming a = b, the linear systems
    # arising in implicit differentiation (if used) of the potentials computed
    # for the symmetric parts should be marked as symmetric.
    kwargs_symmetric.update(
        parallel_dual_updates=True,
        momentum=acceleration.Momentum(start=0, value=0.5),
        anderson=None,
    )
    implicit_diff = kwargs.get("implicit_diff")
    if implicit_diff is not None:
      kwargs_symmetric["implicit_diff"] = implicit_diff.replace(symmetric=True)

  out_xy = linear.solve(geometry_xy, a=a, b=b, **kwargs)
  out_xx = linear.solve(geometry_xx, a=a, b=a, **kwargs_symmetric)
  if geometry_yy is None:
    # Create dummy output, corresponds to scenario where static_b is True.
    out_yy = _empty_output(is_low_rank)
  else:
    out_yy = linear.solve(geometry_yy, a=b, b=b, **kwargs_symmetric)

  eps = jax.lax.stop_gradient(geometry_xy.epsilon)
  div = (
      out_xy.reg_ot_cost - 0.5 * (out_xx.reg_ot_cost + out_yy.reg_ot_cost) +
      0.5 * eps * (jnp.sum(a) - jnp.sum(b)) ** 2
  )

  if is_low_rank:
    factors = tuple((out.q, out.r, out.g) for out in (out_xy, out_xx, out_yy))
    pots = None
  else:
    pots = tuple((out.f, out.g) for out in (out_xy, out_xx, out_yy))
    factors = None

  return SinkhornDivergenceOutput(
      divergence=div,
      geoms=(geometry_xy, geometry_xx, geometry_yy),
      a=a,
      b=b,
      potentials=pots,
      factors=factors,
      errors=(out_xy.errors, out_xx.errors, out_yy.errors),
      converged=(out_xy.converged, out_xx.converged, out_yy.converged),
      n_iters=(out_xy.n_iters, out_xx.n_iters, out_yy.n_iters),
  )


def segment_sinkhorn_divergence(
    x: jnp.ndarray,
    y: jnp.ndarray,
    num_segments: Optional[int] = None,
    max_measure_size: Optional[int] = None,
    cost_fn: Optional[costs.CostFn] = None,
    segment_ids_x: Optional[jnp.ndarray] = None,
    segment_ids_y: Optional[jnp.ndarray] = None,
    indices_are_sorted: bool = False,
    num_per_segment_x: Optional[Tuple[int, ...]] = None,
    num_per_segment_y: Optional[Tuple[int, ...]] = None,
    weights_x: Optional[jnp.ndarray] = None,
    weights_y: Optional[jnp.ndarray] = None,
    solve_kwargs: Mapping[str, Any] = MappingProxyType({}),
    static_b: bool = False,
    share_epsilon: bool = True,
    symmetric_sinkhorn: bool = False,
    **kwargs: Any
) -> jnp.ndarray:
  """Compute Sinkhorn divergence between subsets of vectors in `x` and `y`.

  Helper function designed to compute Sinkhorn divergences between several point
  clouds of varying size, in parallel, using padding for efficiency.
  In practice, The inputs `x` and `y` (and their weight vectors `weights_x` and
  `weights_y`) are assumed to be large weighted point clouds, that describe
  points taken from multiple measures. To extract several subsets of points, we
  provide two interfaces. The first interface assumes that a vector of id's is
  passed, describing for each point of `x` (resp. `y`) to which measure the
  point belongs to. The second interface assumes that `x` and `y` were simply
  formed by concatenating several measures contiguously, and that only indices
  that segment these groups are needed to recover them.

  For both interfaces, both `x` and `y` should contain the same total number of
  segments. Each segment will be padded as necessary, all segments rearranged as
  a tensor, and `vmap` used to evaluate Sinkhorn divergences in parallel.

  Args:
    x: Array of input points, of shape `[num_x, feature]`.
      Multiple segments are held in this single array.
    y: Array of target points, of shape `[num_y, feature]`.
    num_segments: Number of segments contained in `x` and `y`.
      Providing this is required for JIT compilation to work,
      see also :func:`~ott.geometry.segment.segment_point_cloud`.
    max_measure_size: Total size of measures after padding. Should ideally be
      set to an upper bound on points clouds processed with the segment
      interface. Should also be smaller than total length of `x` or `y`.
      Providing this is required for JIT compilation to work.
    cost_fn: Cost function,
      defaults to :class:`~ott.geometry.costs.SqEuclidean`.
    segment_ids_x: **1st interface** The segment ID for which each row of `x`
      belongs. This is a similar interface to :func:`jax.ops.segment_sum`.
    segment_ids_y: **1st interface** The segment ID for which each row of `y`
      belongs.
    indices_are_sorted: **1st interface** Whether `segment_ids_x` and
      `segment_ids_y` are sorted.
    num_per_segment_x: **2nd interface** Number of points in each segment in
      `x`. For example, [100, 20, 30] would imply that `x` is segmented into
      three arrays of length `[100]`, `[20]`, and `[30]` respectively.
    num_per_segment_y: **2nd interface** Number of points in each segment in
      `y`.
    weights_x: Weights of each input points, arranged in the same segmented
      order as `x`.
    weights_y: Weights of each input points, arranged in the same segmented
      order as `y`.
    solve_kwargs: Optionally a dict containing the keywords arguments for
      calls to the `sinkhorn` function, called three times to evaluate for each
      segment the Sinkhorn regularized OT cost between `x`/`y`, `x`/`x`, and
      `y`/`y` (except when `static_b` is `True`, in which case `y`/`y` is not
      evaluated)
    static_b: if True, divergence of measure b against itself is NOT computed
    share_epsilon: if True, enforces that the same epsilon regularizer is shared
      for all 2 or 3 terms of the Sinkhorn divergence. In that case, the epsilon
      will be by default that used when comparing x to y (contained in the first
      geometry). This flag is set to True by default, because in the default
      setting, the epsilon regularization is a function of the mean of the cost
      matrix.
    symmetric_sinkhorn: Use Sinkhorn updates in Eq. 25 of :cite:`feydy:19` for
      symmetric terms comparing x/x and y/y.
    kwargs: keywords arguments passed to form
      :class:`~ott.geometry.pointcloud.PointCloud` geometry objects from the
      subsets of points and masses selected in `x` and `y`, this could be for
      instance entropy regularization float, scheduler or normalization.

  Returns:
    An array of Sinkhorn divergence values for each segment.
  """
  # instantiate padding vector
  dim = x.shape[1]
  if cost_fn is None:
    # default padder
    padding_vector = costs.CostFn._padder(dim=dim)
  else:
    padding_vector = cost_fn._padder(dim=dim)

  def eval_fn(
      padded_x: jnp.ndarray,
      padded_y: jnp.ndarray,
      padded_weight_x: jnp.ndarray,
      padded_weight_y: jnp.ndarray,
  ) -> float:
    mask_x = padded_weight_x > 0.0
    mask_y = padded_weight_y > 0.0
    div, _ = sinkhorn_divergence(
        pointcloud.PointCloud,
        padded_x,
        padded_y,
        a=padded_weight_x,
        b=padded_weight_y,
        solve_kwargs=solve_kwargs,
        static_b=static_b,
        share_epsilon=share_epsilon,
        symmetric_sinkhorn=symmetric_sinkhorn,
        cost_fn=cost_fn,
        src_mask=mask_x,
        tgt_mask=mask_y,
        **kwargs
    )
    return div

  return segment._segment_interface(
      x,
      y,
      eval_fn,
      num_segments=num_segments,
      max_measure_size=max_measure_size,
      segment_ids_x=segment_ids_x,
      segment_ids_y=segment_ids_y,
      indices_are_sorted=indices_are_sorted,
      num_per_segment_x=num_per_segment_x,
      num_per_segment_y=num_per_segment_y,
      weights_x=weights_x,
      weights_y=weights_y,
      padding_vector=padding_vector
  )


def _empty_output(
    is_low_rank: bool
) -> Union[sinkhorn.SinkhornOutput, sinkhorn_lr.LRSinkhornOutput]:
  if is_low_rank:
    return sinkhorn_lr.LRSinkhornOutput(
        q=None,
        r=None,
        g=None,
        ot_prob=None,
        epsilon=None,
        inner_iterations=0,
        converged=True,
        costs=jnp.array([-jnp.inf]),
        errors=jnp.array([-jnp.inf]),
        reg_ot_cost=0.0,
    )

  return sinkhorn.SinkhornOutput(
      potentials=(None, None),
      errors=jnp.array([-jnp.inf]),
      reg_ot_cost=0.0,
      threshold=0.0,
      inner_iterations=0,
  )
